# -*- coding: utf-8 -*-

"""This is the main BUT package.  This provides a basic Brain Use Tactic for use
as an AI layer for a server controlled Khepera wheeled robot in a pseudo-Pong
game."""

import math
from math import pi
import time

from scor.but.khepera import Khepera, DEFENDER, ATTACKER, CATCHER
from .trajectory import Trajectory, Point
from scor.common.config import ScorConfig, ROBOT_SECTION_MASK



class Breakout:

    """BUT AI class -- Breakout category.
    
    This class provides a very basic Pong-mind for a Khepera team playing in a
    SCOR tournament.  All it can do is detect an incoming ball and position
    itself for reception.  In this respect, it is slightly more intelligent than
    the usual Pong AI-controlled player, which would simply aim to stay on the
    same height as the ball.
    
    This class can only defend (for now ?).  Attack strategies are as of now
    unimplemented.  This was the lazy developer's conscious choice.  Really.  It
    was. He thought for hours and all.

    This is a rewrite (the decision loop is rewritten from scratch) of the Pong
    AI."""
    
    def __init__(self, ball, control, cfg, team):
        """
        Initialises the pong object, creating khepera objects
        
        ball : trajectory of the ball
        control : PION communication layer"""
        
        self.ball = ball
        self.control = control
        
        # keeps in memory the last info sent by RIE
        self.last_info_time = None
        self.last_ball_x = None
        self.last_ball_y = None

        self.team = team
        self.kheps = []
        self.terrain_width = cfg.getfloat("terrain", "width")
        self.terrain_length = cfg.getfloat("terrain", "length")
        self.khep_max_speed = cfg.getfloat("robots", "max_speed")

        for robot_id in cfg.robots_ids():
            khep = Khepera(robot_id)
            khep_team = cfg.get(ROBOT_SECTION_MASK % robot_id, "team")
            if khep_team == team:
                self.kheps.append(khep)

        # reads body radius from config
        self.body_radius = cfg.getfloat("robots", "body_radius")

        self._compute_defense_positions()

        self._master_khep = self.kheps[0]

    def update_info(self, info):
        """
        Updates the internal infos with a given set of info

        info : info object returned by the get_info function
        from a coordinate object
        """
        
        # update ball trajectory
        point1_x = self.last_ball_x
        point1_y = self.last_ball_y
        time1 = self.last_info_time
        point2_x = info["ball_coord"][0]
        point2_y = info["ball_coord"][1]
        time2 = info["timestamp"]
        self.ball.update(point1_x, point1_y, time1, point2_x,
            point2_y, time2)
        #update khepera coordinates
        for khep in self.kheps:
            khep.update_coordinates(info)
        # keeps in memory the last info sent by RIE
        self.last_ball_x = info["ball_coord"][0]
        self.last_ball_y = info["ball_coord"][1]
        self.last_info_time = info["timestamp"]


    def _compute_defense_positions(self):
        """
        Calculate and set each robot's (dest_x, dest_y) defense position.  This
        should be called only once, during init, after the team is established
        from the configuration.
        """

        maxid = len(self.kheps)
        khep_pos = 0
        for khep in self.kheps:
            khep_pos = khep_pos + 1
            khep.defense_position = Point(-self.terrain_width/2 + self.terrain_width/(maxid+1)*khep_pos,
                                     -self.terrain_length/4)


    def compute_order(self, khep, dest_x, dest_y):
        """
        # XXX update me
        Return a trajectory order (angle, distance, dest_x, dest_y)
        for a khepera trying to go to dest_x, dest_y.
        """
        
        traj = Trajectory(khep.coord_x, khep.coord_y, time.time(), dest_x, dest_y)
        angle = traj.angle
        distance = math.hypot(khep.coord_x - dest_x, khep.coord_y - dest_y)
        target = Point(dest_x, dest_y)
        return target





    def decide(self):
        """
        The decision-taking function for BUT-Breakout (loop for each khepera).
        """
        elected = None
        if self.ball.incoming:
            elected = self._master_khep
            elected.catch(self.ball)
        elif self.ball.in_our_field and self.ball.speed < self.khep_max_speed:
            elected = self._master_khep
            elected.attack(self.ball)
        else:
            from random import choice
            self._master_khep = choice(self.kheps)
        for khepera in self.kheps:
            if khepera is not elected:
                khepera.defend()
        for khep in self.kheps: # I prefer moving every body when all move are computed
            khep.move(self.control)
