# -*- coding: utf-8 -*-

"""This is the main BUT package.  This provides a basic Brain Use Tactic for use
as an AI layer for a server controlled Khepera wheeled robot in a pseudo-Pong
game."""

import math

from scor.but.khepera import Khepera
from scor.common.config import ScorConfig, ROBOT_SECTION_MASK

class Pong:

    """BUT AI class -- Pong category.
    
    This class provides a very basic Pong-mind for a Khepera team playing in a
    SCOR tournament.  All it can do is detect an incoming ball and position
    itself for reception.  In this respect, it is slightly more intelligent than
    the usual Pong AI-controlled player, which would simply aim to stay on the
    same height as the ball.
    
    This class can only defend (for now ?).  Attack strategies are as of now
    unimplemented.  This was the lazy developer's conscious choice.  Really.  It
    was. He thought for hours and all."""
    
    
    def __init__(self, ball_trajectory, pion_control, cfg, team):
        """
        Initialises the pong object, creating khepera objects
        
        ball_trajectory : trajectory of the ball
        pion_control : PION communication layer"""
        
        self.ball_trajectory = ball_trajectory
        self.pion_control = pion_control
        
        # keeps in memory the last info sent by RIE
        self.last_info_time = None
        self.last_ball_x = None
        self.last_ball_y = None

        self.team = team
        self.kheps = []
        for robot_id in cfg.robots_ids():
            khep = Khepera(robot_id)
            khep_team = cfg.get(ROBOT_SECTION_MASK % robot_id, "team")
            if khep_team == team:
                self.kheps.append(khep)

        self.globalstatus = "dead"
        
        # reads body radius from config
        #self.body_radius = cfg.get("khepera", "body_radius")
        self.body_radius = 0.035
        
        
    def update_info(self, info):
        """
        Updates the internal infos with a given set of info

        info : info object returned by the get_info function
        from a coordinate object
        """
        
        # update ball trajectory
        point1_x = self.last_ball_x
        point1_y = self.last_ball_y
        time1 = self.last_info_time
        point2_x = info["ball_coord"][0]
        point2_y = info["ball_coord"][1]
        time2 = info["timestamp"]
        self.ball_trajectory.update(point1_x, point1_y, time1, point2_x,
            point2_y, time2)
        #update khepera coordinates
        for khep in self.kheps:
            khep.update_coordinates(info)
        # keeps in memory the last info sent by RIE
        self.last_ball_x = info["ball_coord"][0]
        self.last_ball_y = info["ball_coord"][1]
        self.last_info_time = info["timestamp"]

        
    def control_moves(self):
        """
        Controls robots moves
        """
        for khep in self.kheps:

            if khep.status == "dead":
                self.pion_control.rotate(khep.id, 1)
                khep.status = "rotate"

            if khep.status == "rotate":
                if khep.is_pointing_right(khep.order):
                    self.pion_control.stop(khep.id)
                    self.pion_control.move(khep.id, 1)
                    khep.status = "translate"

            if khep.status == "translate":
                if khep.is_at_destination(khep.order):
                    self.pion_control.stop(khep.id)
                    khep.flash(u"["+khep.type+u"] "+u"TARGET ACQUIRED")
                    khep.status = "arrived"
                    
                    if khep.type == "goalkeeper":
                        khep.type = "attacker"
                    elif khep.type == "attacker":
                        khep.type = "walker"

                        
    def decide(self):
        """
        The decision-taking function for BUT-Pong.
        
        BEWARE, I LIVE
        """

        # is the ball coming this way ?
        danger = self.ball_trajectory.is_incoming()

        if (danger and self.globalstatus == "dead"):
            for khep in self.kheps:
                khep_move = \
                    khep.compute_intercepting_move(self.ball_trajectory)
            
                if khep.can_intercept(khep_move, self.ball_trajectory):
                    khep.flash("I OBEY")
                    khep.type = "goalkeeper"
                    self.globalstatus = "goalkeeper"
                    # only 1 khepera can be goalkeeper
                    if khep == self.kheps[1] :
                        khep2 = self.kheps[2]
                    else:
                        khep2 = self.kheps[1]
                    khep2.type = "walker"
                    # Check if khep2 is on khep1 trajectory, or on the trajectory of the ball
                    # when khep1 push it
                    # Coordinate system rotation of angle khep_move["angle"]
                    # Rotated coordinate system is named r2
                    teta = khep_move["angle"]
                    matrix = [[math.cos(teta), math.sin(teta)], [-math.sin(teta), math.cos(teta)]]
                    # inverse rotation matrix
                    phi = - teta
                    inv_matrix = [[math.cos(phi), math.sin(phi)], [-math.sin(phi), math.cos(phi)]]
                    # computes coordinates in the new coordinate system
                    khep1_coord = [khep.pos_x, khep.pos_y]
                    khep1_r2_coord = khep1_coord*matrix
                    khep2_coord = [khep2.pos_x, khep2.pos_y]
                    khep2_r2_coord = khep2_coord*matrix
                    # if the walker is on the goalkeeper trajectory or on the ball trajectory
                    if (khep2_r2_coord[1] > khep1_r2_coord[1]) and \
                        (khep2_r2_coord[2] < 2*self.body_radius) and \
                            (khep2_r2_coord[2] > - 2*self.body_radius):
                        khep2_move = {}
                        if (khep2_r2_coord[2] > 0):
                            # move just out of the way
                            r2_avoiding_move  =  [khep2_r2_coord[1], 2*self.body_radius]
                            avoiding_move = r2_avoiding_move*inv_matrix
                            khep2_move["dest_x"] = avoiding_move[1]
                            khep2_move["dest_y"] = avoiding_move[2]
                            khep2_move["distance"] = 2*self.body_radius - r2_avoiding_move[2]
                            khep2_move["angle"] = teta - math.pi/2
                        elif (khep2_r2_coord[2] < 0):
                            # move just out of the way
                            r2_avoiding_move  =  [khep2_r2_coord[1], - 2*self.body_radius]
                            avoiding_move = r2_avoiding_move*inv_matrix
                            khep2_move["dest_x"] = avoiding_move[1]
                            khep2_move["dest_y"] = avoiding_move[2]
                            khep2_move["distance"] = 2*self.body_radius + r2_avoiding_move[2]
                            khep2_move["angle"] = teta + math.pi/2
                        khep2.order = khep2_move
                        khep2.type = "runner"
                        khep2.status = "dead"
                else:
                    khep.type = "walker"
                khep.status = "dead"
                # TODO: defense trajectories here
                khep.order = khep_move
        return 1
