# -*- coding:utf-8 -*-
"""
This module provides the automagically constructed config parser for all SCOR
modules.
"""

from ConfigParser import SafeConfigParser
import os, sys

from scor.common import defaults
from .defaults.teams import DEFAULT_TEAMS

ROBOT_SECTION_MASK = "robot_%i"

def get_scor_config(extra_config_before=(), extra_config_after=(), argv=None):
    """
    Returns a config parser like object with scor configuration variables, and
    unparsed command line arguments.

    Built-in values and ~/.scor values are used, plus any extra config file
    given.

    Parsing order is the following:
      * built-in config
      * values from extra_config_before
      * values from ~/.scor
      * values from extra_config_after
      * values from argv

    If argv contains the '--config=<path>' option, then path is used instead of
    the default ~/.scor.

    If no argv is provided, sys.argv is used.

    When an argument can't be parsed, it is returned to the caller.
    """
    if argv is None:
        # drops script name, useless
        argv = sys.argv[1:]
    # Try to figure if an alternate home config file should be used
    path, argv = give_defined_config(argv)

    cfg = ScorConfig()
    # first, overrides with extra config files
    for cfg_file in extra_config_before:
        cfg.read_extra(cfg_file)
    # then, reads ~/.scor
    cfg.read_home(path)
    # second round of extra config files
    for cfg_file in extra_config_after:
        cfg.read_extra(cfg_file)
    # finally, parses argv
    unknown_args = cfg.parse_args(argv)
    # finally, returns the ConfigParser created
    return cfg, unknown_args

def give_defined_config(argv):
    """
    Processes given arguments to find whether '--config=<value>' has been
    given (stops at the first found).
    Returns None, argv if it can't be found, otherwise value, argv.
    """
    config = None
    for option in [options for options in argv \
        if options.startswith('--config=')]:
        config = option.split('--config=')[1]
        argv.remove(option)
        break

    return config, argv

class ScorConfig(SafeConfigParser):
    """
    This class lets you easily get all registered configuration values.

    Typical call sequence :
        config = ScorConfig()
        for config_file in my_other_custom_files_with_poney:
            config.read_extra(config_file)
        config.read_home()
        config.parse_args(sys.argv)

    If you are maintaining a module, you should consider using get_scor_config()
    instead, as it makes these calls in the correct order for you (see its
    documentation for more details).
    """

    def __init__(self):
        """
        Creates the configuration parser with built-in configuration variables.
        """
        SafeConfigParser.__init__(self)
        self._read_builtin()

    def _read_builtin(self):
        """
        Reads default configuration variables.
        """
        for module_name, module in defaults.sections.iteritems():
            # Try to avoid referencing private variables
            for variable in [variables for variables in dir(module) \
                if not variables.startswith('_')]:
                # Sections
                if not self.has_section(module_name):
                    self.add_section(module_name)
                # Options
                # This is not very beautiful, but values *must* be strings
                self.set(module_name, variable, str(getattr(module, variable)))

    def read_home(self, alternate_home=None):
        """
        Reads configuration variables contained in alternate_home.
        If alternate_home is not given, ~/.scor is used..
        """
        if alternate_home is None:
            alternate_home = os.path.expanduser("~/.scor")
        self.read(alternate_home)

    def read_extra(self, filename):
        """
        Reads configuration variables from given filename.
        """
        self.read(filename)

    def parse_args(self, argv):
        """
        Reads configuration variables from command line arguments.
        Only existing options can be overriden (no creation).
        Returns all unparsed args.
        """
        variables = {} # {command-line form -> (section, name)}
        unknown_args = []
        # builds allowed command line arguments
        for section in self.sections():
            for option in self.options(section):
                variables['--%s-%s' % (section, option)] = (section, option)

        # process arguments
        for arg in argv:
            splitted_arg = arg.split('=', 1)
            if len(splitted_arg) > 1:
                key, value = splitted_arg
            else:
                key, value = arg, "True"

            section, name = variables.get(key, (None, None))
            if section is not None:
                self.set(section, name, str(value))
            else:
                unknown_args.append(arg)

        return unknown_args

    def getcolor(self, section, option):
        """
        A convenience method which coerces the option in the specified section
        to a tuple.
        """
        val = self.get(section, option)
        # Strips leading and trailing brackets and puts values into a tuple
        return tuple(float(value) for value in val[1:len(val)-1].split(","))

    def robots_ids(self, enabled_only=True):
        """
        Returns the list of all robots defined and (by default) enabled.
        """
        robots = []
        for candidate in self.sections():
            if candidate.startswith("robot_"):
                if enabled_only and self.has_option(candidate, "enabled") \
                                and not self.getboolean(candidate,"enabled"):
                    continue
                robots.append(int(candidate[6:]))
        return robots

    def set_default_teams(self):
        """
        Registers default teams.
        """
        for ident, team, serial in DEFAULT_TEAMS:
            section_name = ROBOT_SECTION_MASK % ident
            if not self.has_section(section_name):
                self.add_section(section_name)
            if not self.has_option(section_name, "team"):
                self.set(section_name, "team", team)
            if not self.has_option(section_name, "port"):
                self.set(section_name, "port", serial)

    def get_robot_infos(self, ident):
        """
        Returns a dictionary containing informations concerning the robot whose
        id is the passed ``ident``.
        """
        return dict(self.items(ROBOT_SECTION_MASK % ident))

# vim: ts=4
