"""Definition of all interface used for intermodule communication"""

class ICoordinate(object):
    """Used to know the position of object on the playfield"""

    def get_info(self, timestamp=None, time_out=None):
        """Result coordinates informations of object on the playfield.

        :type timestamp: number of seconds since epoch (any numeric type)
        :param timestamp: If specified, the call block until data newer
                          than this date is returned

        :type time_out: number max of seconds the call should block
        :param time_out: If specified, the call block until data newer
                          is returned or time out is expire

        :return:
            {
                "timestamp":  data_timestamp,
                "ball_coord": (x, y),
                "khep_coord": {
                                id: (x, y, angle)
                                id: (x, y, angle)
                                id: (x, y, angle)
                                ...
                              },
            }
        :rtype data_timestamp: number of seconds since epoch (any numeric type)
        :rtype x, y: float
        :rtype angle: O rad eastwards, pi/2 rad northwards (towards the
                      opponent's field), modulo 2pi
        """
        raise NotImplementedError()

class IControl(object):
    """
    This class lets you easily control all registered Kheperas using their ids.
    """

    def move(self, id_khepera, speed):
        """
        Starts moving the Khepera associated to id_khepera.

        :type id_khepera: int
        :param id_khepera: The id of the khepera we give order to.
                           If the khepera is unknown a KeyError is raised

        :type speed: float in the [-1, 1] range
        :param speed: Moves forward if speed > 0, backward if speed < 0, and
                      obviously, looks like a sitting duck if speed = 0. 1
                      and -1 are will set the khepera at maximum speed.
        """
        raise NotImplementedError()

    def rotate(self, id_khepera, angular_speed):
        """
        Starts rotate the Khepera associated to id_khepera.

        :type id_khepera: int
        :param id_khepera: The id of the khepera we give order to.
                           If the khepera is unknown a KeyError is raised

        :type speed: float in the [-1, 1] range
        :param speed: rotate clockwise if speed > 0,
                         anticlockwise if speed < 0.
                     -1 and 1 mean maximum speed
        """
        raise NotImplementedError()

    def stop(self, id_khepera):
        """
        Stop the Khepera associated to id_khepera.

        :type id_khepera: int
        :param id_khepera: The id of the khepera we give order to.
                           If the khepera is unknown a KeyError is raised
        """
        raise NotImplementedError()

    def set_speed(self, id_khepera, left_speed, right_speed):
        """
        Set specific speed to each motor to id_khepera.

        :type id_khepera: int
        :param id_khepera: The id of the khepera we give order to.
                           If the khepera is unknown a KeyError is raised
        :type left_speed: float in the [-1, 1] range
        :param left_speed: Moves left wheel forward if speed > 0, backward if
                           speed < 0, and stop it if speed = 0. 1 and -1 are the
                           khepera maximum speed.
        :type right_speed: float in the [-1, 1] range
        :param right_speed: Moves right wheel forward if speed > 0, backward 
                            if speed < 0, and stop it if speed = 0. 
                            1 and -1 are the khepera maximum speed.
        """
        raise NotImplementedError()

class IKheperaSerial(object):
    """
    Serial port used to communicate with a khepera
    """
    #XXX do you need anything else ?
    def read(self, size=1):
        """
        Read size bytes from the serial port. If a timeout is set it may return
        less characters as requested. With no timeout it will block until the
        requested number of bytes is read.
        """
        raise NotImplementedError()
    def write(self, data):
        """
        Output the given string over the serial port.

        :type data: string
        :param data: data you want to write in the serial
        """
        raise NotImplementedError()


class IVideo(object):
    """
    Stream of images comming from a video
    """
    def get_frame(self, timestamp=None, time_out=None):
        """
        Get the most recent image available.
        Block until a new image is found, except if time_out expire
        Return an image newer than timestamp, 
        by default, return a newer image than before

        :type timestamp: number of seconds since epoch (any numeric type)
        :param timestamp: If specified, the call block until image newer
                          than this date is returned

        :type time_out: number max of seconds the call should block
        :param time_out: If specified, the call block until image newer
                          is returned or time out is expire

        :return:
            StringIO of a jpeg image
        """
        raise NotImplementedError()
