# -*- coding:utf-8 -*-

"""This module contains Mock Classes for standard scor interface.

Those objects are intended for testing.
"""

#import ConfigParser
from scor.common.interfaces import IControl
import time

class MockControl(IControl):
    """
    This class lets you easily control all registered Kheperas using their ids.

    This is a variant adapted to provide testing facilities.  Output is done
    through a structure :

    self.khep_list[robot_id][key] = value

    where key is one of "speed", "angular_speed", "left_speed" and "right_speed".

    You can also check a history of calls made via :

    self.callstack[integer][key] = value

    where key is one of "timestamp", "call", "robot_id", and "arglist"

    At no time are actual orders given to actual robots.

    Resemblance to actual people, living or dead, is purely coincidental.

    """

    def __init__(self, robots_ids=(0, 1)):
        """
        Create a fake control layer for two bots.

        # XXX nothing should be parsed !!

        # XXX configuration should be done :

            * either with a scor.common configuration object (usefull for global test)
            * either by direct manipulation of the object (usefull for unit test)

        """
        self._robots_ids = robots_ids
        self.khep_list = {}
        self.callstack = []

        for ident in robots_ids:
            self.khep_list[ident] = {"speed":0, "angular_speed":0, "left_speed":0,
                                     "right_speed":0}


    def move(self, id_khepera, speed):
        """
        Starts moving the Khepera associated to id_khepera if it exists;
        otherwise does nothing.  Moves forward if speed > 0, backward if speed <
        0, and obviously, looks like a sitting duck if speed = 0.
        """
#        print str(id_khepera) + " moves at speed " + str(speed)
        self.khep_list[id_khepera]["left_speed"] = speed
        self.khep_list[id_khepera]["right_speed"] = speed
        self.khep_list[id_khepera]["speed"] = speed
        self.khep_list[id_khepera]["angular_speed"] = 0
        self.callstack.append({"timestamp":time.time(),
                               "call":"move",
                               "robot_id":id_khepera,
                               "arglist":[speed]})


    def rotate(self, id_khepera, angular_speed):
        """
        Start rotating the Khepera associated to id_khepera if it exists;
        otherwise, does nothing.  Rotates clockwise if angular_speed > 0,
        anticlockwise if angular_speed < 0, and starts the "Hare Hare Yukai"
        danse steps if angular_speed = 0 (obviously, that's a lie, but why are
        you reading this documentation anyway?).
        """
#        print str(id_khepera) + " rotates at angular_speed " + str(angular_speed)
        self.khep_list[id_khepera]["left_speed"] = angular_speed
        self.khep_list[id_khepera]["right_speed"] = - angular_speed
        self.khep_list[id_khepera]["angular_speed"] = angular_speed
        self.khep_list[id_khepera]["speed"] = 0
        self.callstack.append({"timestamp":time.time(),
                               "call":"rotate",
                               "robot_id":id_khepera,
                               "arglist":[angular_speed]})


    def stop(self, id_khepera):
        """
        Stops the Khepera associated to id_khepera if it exists; otherwise, does
        nothing.
        """
#        print str(id_khepera) + " stops"
        self.khep_list[id_khepera]["left_speed"] = 0
        self.khep_list[id_khepera]["right_speed"] = 0
        self.khep_list[id_khepera]["angular_speed"] = 0
        self.khep_list[id_khepera]["speed"] = 0
        self.callstack.append({"timestamp":time.time(),
                               "call":"stop",
                               "robot_id":id_khepera,
                               "arglist":[]})


    def set_speed(self, id_khepera, left_speed, right_speed):
        """Separately control each motor.  Does not attempt to set a speed or an angular
        speed."""

#        print str(id_khepera) + " L " + str(left_speed) + " R " + str(right_speed)
        self.khep_list[id_khepera]["left_speed"] = left_speed
        self.khep_list[id_khepera]["right_speed"] = right_speed
        self.callstack.append({"timestamp":time.time(),
                               "call":"set_speed",
                               "robot_id":id_khepera,
                               "arglist":[left_speed, right_speed]})
