# -*- coding:utf-8 -*-
"""This module contains Mock Classes for users of standards scor interface.

Those objects are intended for testing.
"""
from time import sleep

from scor.pion.control import MAX_TICK, TURN_BY_TICK, MOVE_BY_TICK

ANGULAR_FACTOR = MAX_TICK * TURN_BY_TICK
LINEAR_FACTOR = MAX_TICK * MOVE_BY_TICK
#DEG90TIME = 0.25 / ANGULAR_FACTOR
#CM10TIME = 0.1 / LINEAR_FACTOR

FORWARD_MASK = u"%s: forward."
BACKWARD_MASK = u"%s: backward."
ROTATE_CLOCKWISE_MASK = u"%s: rotate clockwise."
ROTATE_ANTICLOCKWISE_MASK = u"%s: rotate anticlockwise."
STOP_MASK = u"%s: stop."
SWITCH = 1

def _stop_on_failure(func):
    """
    Decorator to provide an emergency stop.
    """
    def newfunc(self, ident, delay):
        """
        Provides an emergency stop.
        """
        try:
            func(self, ident, delay)
        except:
            self._stop(ident)
            raise
    return newfunc

class ScorDanseClient(object):
    """Automated client for Control interface."""

    # A class is not really required here but this may be extend to something
    # more powerfull where you can crash your call sequence.

    def __init__(self, control, robots_ids, config): # aaaaaaaaaaaaaaaaaaaaah, nooo
        self.verbose = False
        self.delay = 0
        self.control = control

        # noeeeeeeeeeeeeeeeees
        highest_speed = config.getfloat("robots", "highest_speed")
        max_speed = config.getfloat("robots", "max_speed")

        limit = max_speed/highest_speed

        self.linear_factor  = LINEAR_FACTOR * limit
        self.angular_factor = ANGULAR_FACTOR * limit
        self.robots_ids = sorted(robots_ids)

    def run(self, delay=0, verbose=False):
        """make a sequence of various call to control interface"""

        self.verbose = verbose
        self.delay = delay
        robots_ids = self.robots_ids

        for ident in robots_ids:
            self._stop(ident)

        try:
            self.gimme_a_s(robots_ids[0])
#            self.gimme_a_c(robots_ids[0])
#            self.gimme_an_o(robots_ids[0])
#            self.gimme_an_r(robots_ids[0])
        finally:
            for ident in robots_ids:
                self._stop(ident)

    def gimme_a_s(self, ident):
        """
        Let the robot draw a 'S'.
        """
        try:
            DEG90TIME = 0.25 * self.angular_factor
            CM10TIME =  0.1  * self.linear_factor
            self._forward(ident, CM10TIME) # –
            self._rotate_clockwise(ident, 0.25 ) # ·
            self._backward(ident, CM10TIME) # |
            self._rotate_clockwise(ident, DEG90TIME) # ·
            self._forward(ident, CM10TIME) # –
            self._rotate_anticlockwise(ident, DEG90TIME) # ·
            self._backward(ident, CM10TIME) # |
            self._rotate_anticlockwise(ident, DEG90TIME) # ·
            self._forward(ident, CM10TIME) # –
        finally:
            self._stop(ident)

    def gimme_a_c(self, ident):
        """
        Let the robot draw a 'C'.
        """
        try:
            DEG90TIME = 0.25 * self.angular_factor
            CM10TIME =  0.1  * self.linear_factor
            self._forward(ident, CM10TIME) # –
            self._rotate_anticlockwise(ident, DEG90TIME) # ·
            self._forward(ident, 2 * CM10TIME) # |
            self._rotate_anticlockwise(ident, DEG90TIME) # ·
            self._forward(ident, CM10TIME) # –
        finally:
            self._stop(ident)

    def gimme_an_o(self, ident):
        """
        Let the robot draw an 'O'.
        """
        try:
            DEG90TIME = 0.25 * self.angular_factor
            CM10TIME =  0.1  * self.linear_factor
            self._forward(ident, CM10TIME) # –
            self._rotate_anticlockwise(ident, DEG90TIME) # ·
            self._forward(ident, 2 * CM10TIME) # |
            self._rotate_anticlockwise(ident, DEG90TIME) # ·
            self._forward(ident, CM10TIME) # –
            self._rotate_anticlockwise(ident, DEG90TIME) # ·
            self._forward(ident, 2 * CM10TIME) # |
        finally:
            self._stop(ident)

    def gimme_an_r(self, ident):
        """
        Let the robot draw an 'R'.
        """
        try:
            DEG90TIME = 0.25 * self.angular_factor
            CM10TIME =  0.1  * self.linear_factor
            self._forward(ident, 2*CM10TIME) # |
            self._rotate_clockwise(ident, DEG90TIME) # ·
            self._forward(ident, CM10TIME) # –
            self._rotate_clockwise(ident, DEG90TIME) # ·
            self._forward(ident, CM10TIME) # |
            self._rotate_clockwise(ident, DEG90TIME) # ·
            self._forward(ident, CM10TIME) # –
            self._rotate_anticlockwise(ident, 1.5 * DEG90TIME) # ·
            self._forward(ident, CM10TIME) # \
        finally:
            self._stop(ident)

    def _verbose_print(self, ident, msg_mask):
        """
        Prints a debug statement, only if ``self.verbose=True``.
        """
        if self.verbose:
            print msg_mask % ident

    def _stop(self, ident):
        """
        Stops the robot.
        """
        self._verbose_print(ident, STOP_MASK)
        self.control.stop(ident)

    @_stop_on_failure
    def _forward(self, ident, delay):
        """
        Moves the robot forward for ``delay`` seconds.
        """
        self._verbose_print(ident, FORWARD_MASK)
        self.control.move(ident, SWITCH)
        sleep(delay)

    @_stop_on_failure
    def _backward(self, ident, delay):
        """
        Moves the robot backward for ``delay`` seconds.
        """
        self._verbose_print(ident, BACKWARD_MASK)
        self.control.move(ident, -SWITCH)
        sleep(delay)

    @_stop_on_failure
    def _rotate_clockwise(self, ident, delay):
        """
        Rotates the robot clockwise for ``delay`` seconds.
        """
        self._verbose_print(ident, ROTATE_CLOCKWISE_MASK)
        self.control.rotate(ident, SWITCH)
        sleep(delay)

    @_stop_on_failure
    def _rotate_anticlockwise(self, ident, delay):
        """
        Rotates the robot anticlockwise for ``delay`` seconds.
        """
        self._verbose_print(ident, ROTATE_ANTICLOCKWISE_MASK)
        self.control.rotate(ident, -SWITCH)
        sleep(delay)

# vim: ts=4
