"""
Test cases for the default configuration.
"""

from unittest import TestCase, main as unittest_main
import os, filecmp

from ConfigParser import NoSectionError

from scor.common import config
from scor.common.config import ScorConfig

def write_fake(filename, alternate=False):
    """
    Writes a fake config file with some values.
    """
    params = {}
    # overrides an option
    params[u"general"] = {
        u"nb_kheperas" : u"750"
    }
    # adds an option into an existing section
    params[u"const"] = {
        u"zomg_ponies" : u"lol"
    }
    # new section with two new options
    params[u"this_section_does_not_exist"] = {
        u"yes_it_exists" : u"no",
        u"no_it_does_not_exist" : u"yes"
    }
    if alternate:
        params[u"this_section_exists"] = {
            u"yes_it_exists" : u"no",
            u"no_it_does_not_exist" : u"yes"
        }
    fout = open(filename, u"w")
    for section, options in params.items():
        fout.write(u"[" + section + u"]\n")
        for key, value in options.items():
            fout.write(key + u"=" + value + u"\n")
    fout.close()

class ScorConfigTC(TestCase):

    def test_robots_ids_empty(self):
        config = ScorConfig()
        self.assertEquals(config.robots_ids(), [])

    def test_robots_ids_with_robots(self):
        config = ScorConfig()
        config.add_section("robot_1")
        config.add_section("robot_2")
        config.add_section("robot_42")
        self.assertEquals(set(config.robots_ids()), set([1, 2, 42]))

    def test_robots_ids_without_robots(self):
        config = ScorConfig()
        config.add_section("toto")
        config.add_section("bob_1337")
        self.assertEquals(config.robots_ids(), [])

    def test_robots_ids_enabled_only(self):
        config = ScorConfig()
        config.add_section("robot_1")
        config.add_section("robot_2")
        config.set("robot_2", "enabled", "0")
        config.add_section("robot_42")
        self.assertEquals(set(config.robots_ids()), set([1, 42]))

    def test_robots_ids_all(self):
        config = ScorConfig()
        config.add_section("robot_1")
        config.add_section("robot_2")
        config.set("robot_2", "enabled", "0")
        config.add_section("robot_42")
        self.assertEquals(set(config.robots_ids(enabled_only=False)), set([1, 2, 42]))

    def test_default_teams_absent(self):
        config = ScorConfig()
        self.assertEquals(config.robots_ids(), [])

    def test_default_teams_present(self):
        config = ScorConfig()
        config.set_default_teams()
        self.assertEquals(set(config.robots_ids()), set([1, 2, 3, 4]))

    def test_default_teams_team_values(self):
        config = ScorConfig()
        config.set_default_teams()

        self.assertTrue(config.has_option("robot_2", "team"))
        self.assertEquals(config.get("robot_2","team"), "south")
        self.assertTrue(config.has_option("robot_3", "team"))
        self.assertEquals(config.get("robot_3","team"), "north")

    def test_default_teams_serial_values(self):
        config = ScorConfig()
        config.set_default_teams()

        self.assertTrue(config.has_option("robot_1", "port"))
        self.assertTrue(config.has_option("robot_3", "port"))
        self.assertNotEquals(config.get("robot_1","port"),
                             config.get("robot_3","port"))

    def test_default_teams_no_overwrite(self):
        config = ScorConfig()
        config.add_section("robot_1")
        config.set("robot_1", "port", "tardis")
        config.set_default_teams()

        self.assertEquals(config.get("robot_1","port"), "tardis")

    def test_get_robot_infos_absent(self):
        config = ScorConfig()
        self.assertRaises(NoSectionError, config.get_robot_infos, 0)

    def test_get_robot_infos_values(self):
        config = ScorConfig()
        config.add_section(u"robot_2")
        config.set(u"robot_2", u"enabled", u"0")

        self.assertEquals(config.get_robot_infos(2), dict([(u"enabled", u"0")]))

class DefaultConfig_TestCase(TestCase):
    """
    Runs tests on the default configuration interface.
    """

    def setUp(self):
        """
        Creates a default configuration.
        """
        self.datadir = u"data"
        self.references = [os.path.join(self.datadir, u"builtin.ini"), \
            os.path.join(self.datadir, u"builtin+home.ini"), \
            os.path.join(self.datadir, u"builtin+extra.ini"), \
            os.path.join(self.datadir, u"builtin+args.ini"), \
            os.path.join(self.datadir, u"full+alternate.ini"), \
            os.path.join(self.datadir, u"full.ini")]
        self.filenames = [os.path.join(self.datadir, u"lol0.ini"), \
            os.path.join(self.datadir, u"lol1.ini"), \
            os.path.join(self.datadir, u"lol2.ini"), \
            os.path.join(self.datadir, u"lol3.ini"), \
            os.path.join(self.datadir, u"lol4.ini"), \
            os.path.join(self.datadir, u"lol5.ini")]
        self.home = os.path.expanduser("~/.scor")
        self.home_backup = "%s.%s" % (self.home, u"bak")
        self.cfg = ScorConfig()
        self.index = None

    def tearDown(self):
        """
        Deletes the default config and the file written to disk.
        """
        del self.cfg

        for cfg_file in [self.filenames[self.index], self.home]:
            if os.path.exists(cfg_file):
                os.remove(cfg_file)
        if os.path.exists(self.home_backup):
            os.rename(self.home_backup, self.home)

    def test_builtin(self):
        """
        Checks that parsed configuration values match pristine ones.
        """
        self.index = 0
        self.assertTrue(self._compare_config(self.references[self.index], \
            self.filenames[self.index]))

    def test_getcolor(self):
        """
        Tests that getcolor() method returns correct colors (3 comp + 1 alpha).
        """
        self.index = 0
        self.assertEqual((0.0, 0.0, 0.0, 1.0), \
            self.cfg.getcolor(u"terrain", u"border_material_color"))

    def test_home(self):
        """
        Checks adding and overriding default values with ~/.scor.
        """
        self.index = 1
        # saves ~/.scor
        if os.path.exists(self.home):
            os.rename(self.home, self.home_backup)
        # writes ~/.scor
        write_fake(self.home)
        # reads it
        self.cfg.read_home()
        self.assertTrue(self._compare_config(self.references[self.index], \
            self.filenames[self.index]))

    def test_extra(self):
        """
        Checks adding and overriding default values with an extra file.
        """
        self.index = 2
        self.cfg.read_extra(self.references[self.index-1])
        self.assertTrue(self._compare_config(self.references[self.index], \
            self.filenames[self.index]))

    def test_parse(self):
        """
        Checks overriding defaults values with the command line.
        """
        self.index = 3
        argv = ['--ball-x=1', '--foo', '--bar', \
            '--terrain-middle_wall_material_image=chaussette.png', '--baz']
        # --foo, --bar and --baz should be returned
        self.assertEqual(self.cfg.parse_args(argv), ['--foo', '--bar', '--baz'])
        self.assertTrue(self._compare_config(self.references[self.index], \
            self.filenames[self.index]))

    def test_get(self):
        """
        Tests get_scor_config() with an alternate home config file.
        """
        self.index = 4
        # saves .scor
        if os.path.exists(u".scor"):
            os.rename(u".scor", u".scor.bak")
        # writes ~/.scor
        write_fake(u".scor", True)
        # Reads it
        tmp_config, unknown_args = config.get_scor_config((), (),
            ['--config=.scor', '--config=pouet'])
        self.assertEqual(unknown_args, ['--config=pouet'])
        fout = open(self.filenames[self.index], u"w")
        tmp_config.write(fout)
        fout.close()
        # some cleaning
        del tmp_config
        if os.path.exists(u".scor"):
            os.remove(u".scor")
        if os.path.exists(u".scor.bak"):
            os.rename(u".scor.bak", u".scor")
        self.assertTrue(filecmp.cmp(self.references[self.index], \
            self.filenames[self.index], 0))

    def test_get_without_arg(self):
        """
        Tests get_scor_config() without command line arguments.
        """
        self.index = 5
        # saves ~/.scor
        if os.path.exists(self.home):
            os.rename(self.home, self.home_backup)
        # writes ~/.scor
        write_fake(self.home)
        # Reads it
        tmp_config, unknown_args = config.get_scor_config(
            (self.references[self.index-3],), (self.references[self.index-2],))
        self.assertEqual(unknown_args, [])
        fout = open(self.filenames[self.index], u"w")
        tmp_config.write(fout)
        fout.close()
        # some cleaning
        del tmp_config
        self.assertTrue(filecmp.cmp(self.references[self.index], \
            self.filenames[self.index], 0))

    def _compare_config(self, reference, filename):
        """
        Writes the config to a temporary file, and compares it with the
        reference one.
        """
        fout = open(filename, u"w")
        self.cfg.write(fout)
        fout.close()
        return filecmp.cmp(reference, filename, 0)

if __name__ == '__main__':
    unittest_main()

# vim: ts=4
