# -*- coding:utf-8 -*-
"""
This module contains the control interface for all Khepera registered in SCOR.

If a Khepera isn't registered in the config file, you won't be able to control
it!
"""
import Pyro.core

from scor.common.interfaces import IControl

# Harcoding Khep params for now
MAX_TICK = 127
TURN_BY_TICK = 0.048
MOVE_BY_TICK = 0.008
MOVE_MASK = 'D,%i,%i\n'


def send_speed(serial_like, left_engine, right_engine):
    """
    Sets engines' speed (left engine speed is set to left_engine and
    right engine one to right_engine).
    """
    serial_like.write(MOVE_MASK % (left_engine, right_engine))

class Control(IControl, Pyro.core.ObjBase):
    """
    This class lets you easily control all registered Kheperas using their ids.
    """

    def __init__(self, khep_list, limit):
        """
        Registers controllable Kheperas.
        """
        IControl.__init__(self)
        Pyro.core.ObjBase.__init__(self)
        # Here they are
        self._khep_list = khep_list
        self.limit = limit
        print limit

    def move(self, id_khepera, speed):
        """
        Starts moving the Khepera associated to id_khepera.

        :type id_khepera: int
        :param id_khepera: The id of the khepera we give order to.
                           If the khepera is unknown a KeyError is raised

        :type speed: float in the [-1, 1] range
        :param speed: Moves forward if speed > 0, backward if speed < 0, and
                      obviously, looks like a sitting duck if speed = 0. 1
                      and -1 are will set the khepera at maximum speed.
        """
        real_speed = self.translate_speed(speed)
        send_speed(self._khep_list[id_khepera], real_speed, real_speed)

    def rotate(self, id_khepera, angular_speed):
        """
        Starts rotate the Khepera associated to id_khepera.

        :type id_khepera: int
        :param id_khepera: The id of the khepera we give order to.
                           If the khepera is unknown a KeyError is raised

        :type speed: float in the [-1, 1] range
        :param speed: rotate clockwise if speed > 0,
                         anticlockwise if speed < 0.
                     -1 and 1 mean maximum speed
        """
        real_speed = self.translate_speed(angular_speed)
        send_speed(self._khep_list[id_khepera], real_speed, -real_speed)

    def stop(self, id_khepera):
        """
        Stop the Khepera associated to id_khepera.

        :type id_khepera: int
        :param id_khepera: The id of the khepera we give order to.
                           If the khepera is unknown a KeyError is raised
        """
        send_speed(self._khep_list[id_khepera], 0, 0)

    def set_speed(self, id_khepera, left_speed, right_speed):
        """
        Set specific speed to each motor to id_khepera.

        :type id_khepera: int
        :param id_khepera: The id of the khepera we give order to.
                           If the khepera is unknown a KeyError is raised
        :type left_speed: float in the [-1, 1] range
        :param left_speed: Moves left wheel forward if speed > 0, backward if
                           speed < 0, and stop it if speed = 0. 1 and -1 are the
                           khepera maximum speed.
        :type right_speed: float in the [-1, 1] range
        :param right_speed: Moves right wheel forward if speed > 0, backward-
                            if speed < 0, and stop it if speed = 0.-
                            1 and -1 are the khepera maximum speed.
        """
        left_speed = self.translate_speed(left_speed)
        right_speed = self.translate_speed(right_speed)
        send_speed(self._khep_list[id_khepera], left_speed, right_speed)

    def translate_speed(self, speed, limit=None):
        """
        Translates given speed to a value relevant to the khepera.

        Does not consider only -1 and 1, but all floating values between, as
        this will be useful later, when EDNA is ready.
        """
        if limit is None:
            limit = self.limit
        if speed > 0:
            return min(MAX_TICK*limit, int(MAX_TICK*limit*speed))
        else:
            return max(-MAX_TICK*limit, int(MAX_TICK*limit*speed))

# vim: ts=4
