# -*- coding:utf-8 -*-
"""
This module is responsible for managing the Kheperas with their control
interface.
"""
import Pyro.core
import serial

from scor.common.config import get_scor_config
from scor.pion.control import Control
from scor.pion.wrapper import Wrapper, SerialHub

class Manager(object):
    """
    This class registers Kheperas.
    """

    def __init__(self):
        """
        Parses the config file and instantiates needed Kheperas.
        """
        self._ctrl = None
        self._pyro_server = None

        # Config files parsing
        self.config, argv = get_scor_config()
        # Create some players/teams by default
        self.config.set_default_teams()

        # Real Kheperas or not?
        self._simulation = False
        self._debug = False
        self._reality = False
        for option in argv:
            if option.startswith('--simulate'):
                print "serial mode : simulated"
                self._simulation = True
            elif option.startswith('--debug'):
                print "serial mode : debug"
                self._debug = True
            elif option.startswith('--real'):
                print "serial mode : real"
                self._reality = True
        if not (self._simulation or self._debug or self._reality):
            print "serial mode : real"
            self._reality = True

        # Instanciate a control interface with registered Kheperas
        self._create_control(self.get_kheps())

    def start(self):
        """
        Starts the manager (actually, it only starts Pyro's main loop).
        """
        try:
            # Start pyro server
            self._pyro_server.requestLoop()
        finally:
            self._pyro_server.shutdown(True)

    def _create_control(self, khep_list):
        """
        Creates the control interface and exposes it through Pyro.
        """
        highest_speed = self.config.getfloat("robots", "highest_speed")
        max_speed = self.config.getfloat("robots", "max_speed")

        self._ctrl = Control(khep_list, limit=max_speed/highest_speed)

        # Load config
        server_host = self.config.get(u"pyro_server", u"control_host")
        server_port = self.config.getint(u"pyro_server", u"control_port")
        # Init Pyro server
        Pyro.core.initServer()
        self._pyro_server = Pyro.core.Daemon(host=server_host, port=server_port)
        self._pyro_server.connect(self._ctrl, u"control")
        print "Pyro server bound on '%s:%i'." % (server_host, server_port)

    def get_kheps(self):
        """
        Returns a dictionary containing (id, serial) couples.
        """
        khep_list = {}

        if self._simulation:
            # Load config
            serial_host = self.config.get(u"pyro_server", u"serial_host")
            serial_port = self.config.getint(u"pyro_server", u"serial_port")
            # Init Pyro client
            Pyro.core.initClient()

        # Create them
        for i in self.config.robots_ids(True):
            serial_like = None
            port = self.config.get_robot_infos(i)[u"port"]
            # Simulation
            serial_like = []
            if self._simulation:
                # Build the uri
                uri = "PYROLOC://%s:%i/serial/%s" % (serial_host, serial_port,
                    port)
                # Get the remote object
                serial_like.append(Pyro.core.getProxyForURI(uri))
            # Debug
            if self._debug:
                serial_like.append(Wrapper(port))
            # Reality
            if self._reality:
                serial_like.append(serial.Serial(port, 9600, serial.EIGHTBITS,
                    serial.PARITY_NONE, serial.STOPBITS_TWO, timeout=0))

            # And, finally, register it :)
            assert serial_like
            if len(serial_like) > 1:
                serial_like = SerialHub(serial_like)
            else:
                serial_like = serial_like[0]

            khep_list[i] = serial_like

        return khep_list

# vim: ts=4
