#!/usr/bin/env python
# pylint: disable-msg=W0404,W0622,W0704,W0613,W0403,C0103
"""Generic Setup script, takes package info from __pkginfo__.py file.

:copyright: 2003-2008 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
:license: General Public License version 2 - http://www.gnu.org/licenses
"""
__docformat__ = "restructuredtext en"

import os
import sys
import shutil
from distutils.core import setup
from distutils.command import build
from distutils.command import install_lib
from os.path import isdir, exists, join, walk

# import required features
from __pkginfo__ import modname, version, license, short_desc, long_desc, \
     web, author, author_email
# import optional features
try:
    from __pkginfo__ import distname #pylint: disable-msg=E0611
except ImportError:
    distname = modname
try:
    from __pkginfo__ import scripts #pylint: disable-msg=E0611
except ImportError:
    scripts = []
try:
    from __pkginfo__ import data_files #pylint: disable-msg=E0611
except ImportError:
    data_files = None
try:
    from __pkginfo__ import subpackage_of #pylint: disable-msg=E0611
except ImportError:
    subpackage_of = None
try:
    from __pkginfo__ import include_dirs #pylint: disable-msg=E0611
except ImportError:
    include_dirs = []
try:
    from __pkginfo__ import ext_modules #pylint: disable-msg=E0611
except ImportError:
    ext_modules = None

STD_BLACKLIST = ('CVS', '.svn', '.hg', 'debian', 'dist', 'build')

IGNORED_EXTENSIONS = ('.pyc', '.pyo', '.elc', '~')

    

def ensure_scripts(linux_scripts):
    """
    Creates the proper script names required for each platform
    (taken from 4Suite)
    """
    from distutils import util
    if util.get_platform()[:3] == 'win':
        scripts_ = [script + '.bat' for script in linux_scripts]
    else:
        scripts_ = linux_scripts
    return scripts_


def get_packages(directory, prefix):
    """return a list of subpackages for the given directory
    """
    result = []
    for package in os.listdir(directory):
        absfile = join(directory, package)
        if isdir(absfile):
            if exists(join(absfile, '__init__.py')) or \
                   package in ('test', 'tests'):
                if prefix:
                    result.append('%s.%s' % (prefix, package))
                else:
                    result.append(package)
                result += get_packages(absfile, result[-1])
    return result

def export(from_dir, to_dir,
           blacklist=STD_BLACKLIST,
           ignore_ext=IGNORED_EXTENSIONS):
    """make a mirror of from_dir in to_dir, omitting directories and files
    listed in the black list
    """
    def make_mirror(arg, directory, fnames):
        """walk handler"""
        for norecurs in blacklist:
            try:
                fnames.remove(norecurs)
            except ValueError:
                pass
        for filename in fnames:
            # don't include binary files
            if filename[-4:] in ignore_ext:
                continue
            if filename[-1] == '~':
                continue
            src = '%s/%s' % (directory, filename)
            dest = to_dir + src[len(from_dir):]
            print >> sys.stderr, src, '->', dest
            if os.path.isdir(src):
                if not exists(dest):
                    os.mkdir(dest)
            else:
                if exists(dest):
                    os.remove(dest)
                shutil.copy2(src, dest)
    try:
        os.mkdir(to_dir)
    except OSError, ex:
        # file exists ?
        import errno
        if ex.errno != errno.EEXIST:
            raise
    walk(from_dir, make_mirror, None)


EMPTY_FILE = '"""generated file, don\'t modify or your data will be lost"""\n'

class BuildScripts(install_lib.install_lib): # pylint: disable-msg=R0904

    def run(self):
        install_lib.install_lib.run(self)


        if include_dirs:
            base = modname
            for directory in include_dirs:
                dest = join(self.install_dir, base, directory)
                export(directory, dest)
        
class Build(build.build): # pylint: disable-msg=R0904
    """Extended Build Command creating __init__.py for subpackage"""

    def run(self):
        """execute actual build process"""
        build.build.run(self)

        #createparent package __init__.py
        install_path = self.build_lib
        for parent_pkg in subpackage_of.split('.'):
            install_path = join(install_path, parent_pkg)
            init_path = join(install_path, '__init__.py')
            if not exists(init_path):
                self.announce('creating %s __init__.py' % parent_pkg)
                if not self.dry_run:
                    if not exists(install_path):
                        os.makedirs(install_path)
                    stream = open(init_path, 'w')
                    stream.write(EMPTY_FILE)
                    stream.close()

def install(**kwargs): # pylint: disable-msg=W0142
    """setup entry point"""
    if subpackage_of:
        package = subpackage_of + '.' + modname
        kwargs['package_dir'] = {package : '.'}
        packages = [package] + get_packages(os.getcwd(), package)

    else:
        kwargs['package_dir'] = {modname : '.'}
        packages = [modname] + get_packages(os.getcwd(), modname)
    kwargs['packages'] = packages
    return setup(name = distname,
                 version = version,
                 license =license,
                 description = short_desc,
                 long_description = long_desc,
                 author = author,
                 author_email = author_email,
                 url = web,
                 scripts = ensure_scripts(scripts),
                 data_files=data_files,
                 ext_modules=ext_modules,
                 cmdclass={'install_lib': BuildScripts,
                           'build': Build},
                 **kwargs
                 )
            
if __name__ == '__main__' :
    install()
