# -*- coding:utf-8 -*-
"""Handy creation of the faces of a cylinder"""
# pylint: disable-msg=R0913,R0914

from math import sin, cos, pi
from soya import World, Face, Vertex

def Cylinder(parent = None, length=1, radius=1,  #pylint: disable-msg=C0103
        material = None, slices = 20, smooth_lit = 1, insert_into = None):
    """
    Creates a cylinder.

    Creates and returns a World in PARENT, containing a cylinder of specified
    radius and lenght, centered on the origin, and aligned along the local Z
    axis, with material MATERIAL. The material is mapped on the three faces.

    SLICES can be used to control the quality of the cylinder.

    If INSERT_INTO is not None, the cylinder's faces are inserted into it,
    instead of creating a new world.
    """

    radius = float(radius)
    milength = float(length) / 2
    
    cylinder = insert_into or World(parent)
    
    step = 2 * pi / slices
    
    angle = 0.0

    # The centers of the two faces
    center_pos = Vertex(cylinder, 0, 0, milength, 0.75, 0.25)
    center_neg = Vertex(cylinder, 0, 0, -milength, 0.25, 0.25)

    for i in xrange(slices): #pylint: disable-msg=W0612
        # Creation of the two disks
        current_pos = Vertex(cylinder, radius * cos(angle),
                                       radius * sin(angle),
                                       milength,
                                       0.75 - sin(angle)/4,
                                       0.25 - cos(angle)/4)
        current_neg = Vertex(cylinder, radius * cos(angle),
                                       radius * sin(angle),
                                       -milength,
                                       0.25 - sin(-angle)/4,
                                       0.25 - cos(-angle)/4)
        next_pos = Vertex(cylinder, radius * cos(angle + step), 
                                    radius * sin(angle + step), 
                                    milength,
                                    0.75 - sin(angle + step)/4,
                                    0.25 - cos(angle + step)/4)
        next_neg = Vertex(cylinder, radius * cos(angle + step), 
                                    radius * sin(angle + step), 
                                    -milength,
                                    0.25 - sin(-angle - step)/4,
                                    0.25 - cos(-angle - step)/4)
        Face(cylinder, [center_pos, current_pos, next_pos], material)
        Face(cylinder, [center_neg,  next_neg, current_neg], material)

        # Creation of the central faces
        current_pos = Vertex(cylinder, radius * cos(angle),
                                       radius * sin(angle),
                                       milength,
                                       float(i) / slices,
                                       0.5)
        current_neg = Vertex(cylinder, radius * cos(angle),
                                       radius * sin(angle),
                                       -milength,
                                       float(i) / slices,
                                       1)
        next_pos = Vertex(cylinder, radius * cos(angle + step), 
                                    radius * sin(angle + step), 
                                    milength,
                                    float(i + 1) / slices,
                                    0.5)
        next_neg = Vertex(cylinder, radius * cos(angle + step), 
                                    radius * sin(angle + step), 
                                    -milength,
                                    float(i + 1) / slices,
                                    1)
        face_length = Face(cylinder, [current_pos, 
                                      current_neg, 
                                      next_neg, 
                                      next_pos],
                                      material)
        face_length.smooth_lit = smooth_lit
        angle += step
        
    return cylinder
