#!/usr/bin/env python
# -*- coding:utf-8 -*-
#pylint: disable-msg=W0233,W0231,E1101
""" DragDropWorld is a world that allows to dragdrop its content with the mouse.
"""

import soya, soya.sdlconst
from scor.sonere import models

class DragDropWorld(soya.World):
    """
    DragDropWorld is a world that allows to dragdrop its content with the
    mouse.
    """

    FORCE_COEFF = 5

    def __init__(self):
        """
        Initialize attributes.

        Extends soya.World.__init__
        """
        soya.World.__init__(self)
        
        self.terrain = None
        # The object we are currently dragdroping (None => no dragdrop).
        self.dragdroping = None
        
        # The impact point
        self.impact = None

        # The camera
        self.camera = None

        self.old_mouse = None
        self._force = None
        
    def set_camera(self, camera):
        """
        Set the camera attribute.
        """
        self.camera = camera

    def process_events(self):
        """
        Perform drag and drop.
        """
        # Processes the events
        for event in soya.MAIN_LOOP.events:
            # Mouse down initiates the dragdrop.
            if   event[0] == soya.sdlconst.MOUSEBUTTONDOWN:

                # Default fonction in drag and drop is a move
                self._force = False
                
                # The event give us the 2D mouse coordinates in pixel. The
                # camera.coord2d_to_3d convert these 2D pixel coordinates into
                # a soya.Point object.
                mouse = self.camera.coord2d_to_3d(event[2], event[3])
                
                # Performs a raypicking, starting at the camera and going
                # toward the mouse.
                # The vector_to method returns the vector between 2 positions.
                # This raypicking grabs anything that is under the mouse.
                # Raypicking returns
                # None if nothing is encountered, or a (impact, normal) tuple,
                # where impact is the position of the impact and normal is the
                # normal vector at this position.
                # The object encountered is impact.parent ; here, we don't need
                # the normal.
                result = self.raypick(self.camera,
                    self.camera.vector_to(mouse))
                if result:
                    self.impact, normal = result #pylint: disable-msg=W0612
                    # parent**2 to get the Khepera world, else it's the ball 
                    self.dragdroping = self.impact.parent
                    if not isinstance(self.dragdroping, models.Ball):
                        self.dragdroping = self.dragdroping.parent
                    else:
                        # It's the Ball
                        if event[1] == soya.sdlconst.BUTTON_RIGHT:
                            # We apply a new force to the ball instead of
                            # moving it
                            self._force = True
                        else:
                            # We drop all the force applyed to the ball
                            self.dragdroping.angular_velocity = soya.Vector(None, 0, 0, 0) #pylint: disable-msg=C0301
                            self.dragdroping.linear_velocity  = soya.Vector(None, 0, 0, 0) #pylint: disable-msg=C0301
                    
                    # Converts impact into the camera coordinate system, in
                    # order to get its Z value.
                    # camera.coord2d_to_3d cannot choose a Z value for you, so
                    # you need to pass it as a third argument (it defaults to
                    # -1.0). Then, we computes the old mouse position, which
                    # has the same Z value than impact.
                    self.impact.convert_to(self.camera)
                    self.old_mouse = self.camera.coord2d_to_3d(event[2],
                        event[3], self.impact.z)
                    
            # Mouse up ends the dragdrop.
            elif event[0] == soya.sdlconst.MOUSEBUTTONUP:
                if self._force:
                    # We apply the force now
                    mouse = self._convert_coord(event[2], event[3])
                    force = self.dragdroping.vector_to(mouse) * self.FORCE_COEFF
                    self.dragdroping.add_force(force)
                    self._force = False
                self.dragdroping = None
                
            # Mouse motion moves the dragdroping object, if there is one.
            elif event[0] == soya.sdlconst.MOUSEMOTION:
                if self.dragdroping and not self._force:
                    # Computes the new mouse position, at the same Z value than
                    # impact.
                    new_mouse = self._convert_coord(event[1], event[2])

                    # Translates dragdroping by a vector starting at old_mouse
                    # and ending at new_mouse.
                    delta = self.old_mouse.vector_to(new_mouse)
                    self.dragdroping.add_vector(delta)
                    
                    # Store the current mouse position.
                    self.old_mouse = new_mouse



    def _convert_coord(self, x, y): #pylint: disable-msg=C0103
        """
        Convert a 3D coordinate to a 2D in field coordinate.
        Make a correction of needed.
        """
        coord_3d = self.camera.coord2d_to_3d(x, y, self.impact.z)
        coord_2d = self._coord3d_to_field(coord_3d)
        if not self._force:
            # We want not to move the object out of the field
            coord_2d = self._coord_correction(coord_2d)
        return coord_2d

    def _coord3d_to_field(self, coord_3d):
        """
        Convert a 3d coordinate to a 2d coordinate ini the field
        """
        if self.terrain:
            # in the Field's coordinates
            coord_3d = coord_3d % self.terrain
            # TODO: à changer avec les bonnes coordonnées
            # We apply a correction to compensate the Y mouse
            # coordinate
            correction = self.camera.vector_to(coord_3d) % self
            correction.normalize()
            x_corr = coord_3d.y * correction.x
            z_corr = coord_3d.y * correction.z
            coord_3d.x += x_corr
            coord_3d.y = 0.03
            coord_3d.z += z_corr
        return coord_3d

    def _coord_correction(self, coord_2d):
        """
        Don't go over the field's border
        """
        xmax = self.terrain.width/2 - 0.01
        zmax = self.terrain.length/2 - 0.01
        if abs(coord_2d.x) > xmax:
            coord_2d.x = coord_2d.x > 0 and xmax or -xmax
        if abs(coord_2d.z) > zmax:
            coord_2d.z = coord_2d.z > 0 and zmax or -zmax
        return coord_2d
