#!/usr/bin/env python
# -*- coding:utf-8 -*-
# pylint: disable-msg=W0231
"""
Implementation of communication interfaces.
"""

from __future__ import division
from time import time
from math import pi
import Pyro.core
import threading
from StringIO import StringIO

from scor.common.interfaces import IControl, ICoordinate, IKheperaSerial, IVideo

class Control(IControl):
    """
    This class lets you easily control all registered Kheperas using their ids.
    """
    MOVE_COMMAND = u"D"

    def __init__(self, kheperas_dict):
        """
        Initialize the control object with khepera identified in kheperas_dict.
        """
        self._kheperas = kheperas_dict

    def move(self, id_khepera, speed):
        """
        Starts moving the Khepera associated to id_khepera if it exists;
        otherwise does nothing.
        Moves forward (at max speed) if speed = 1, backward if speed = -1,
        and stop if speed = 0.
        speed must be between -1 and 1.

        """
        #print "move : speed=", speed, " idKhep=", id_khepera
        self.set_speed(id_khepera, speed, speed)

    def rotate(self, id_khepera, angular_speed):
        """
        Start rotating the Khepera associated to id_khepera if it exists;
        otherwise, does nothing.
        Rotates clockwise if angular_speed = 1, 
        anticlockwise if angular_speed = -1,
        and stop if angular_speed = 0.
        angular_speed must be between -1 and 1.
        """
        #print "rotate : angular_speed=", angular_speed, " idKhep=", id_khepera
        self.set_speed(id_khepera, angular_speed, -angular_speed)

    def stop(self, id_khepera):
        """
        Stops the Khepera associated to id_khepera if it exists; otherwise, does
        nothing.
        """
        #print "stop : idKhep=", id_khepera
        self.set_speed(id_khepera, 0, 0)

    def set_speed(self, id_khepera, left_speed, right_speed):
        """
        Set the speed separatly for each motor associated to id_khepera if it exists; otherwise does nothing.
        """
        if self._kheperas.has_key(id_khepera):
            self._kheperas[id_khepera].new_order(left_speed, right_speed)

    def switch_light(self, id_khepera, switch):
        """
        Switch on the light (the LED above the Khepera) : set it visible.

        @type   id_khepera: number
        @param  id_khepera: the id of the khepera to be controlled
        @type   switch: boolean
        @param  switch: True to switch on the light, False to switch it off
        """
        if self._kheperas.has_key(id_khepera):
            self._kheperas[id_khepera].led1.visible = switch
            self._kheperas[id_khepera].led_timer = self._kheperas[id_khepera].TIMER



class Coordinate(ICoordinate):
    """
    This class provides easy access to positions of kheperas and ball.
    """
    TIMESTAMP = u"timestamp"
    BALL      = u"ball_coord"
    KHEPERAS  = u"khep_coord"

    # Constant to set the team name.
    NORTH = -1
    SOUTH =  1

    def __init__(self, terrain, ball, kheperas, team):
        """
        Initialize attribute of the object.

        terrain: the Terrain
        ball: the Ball
        kheperas: dict of {num => Khepera}
        team: team name. Must be Coordinate.NORTH or Coordinate.SOUTH
        """
        self.terrain = terrain
        self.kheperas = kheperas
        self.ball = ball
        self.team = team

        self.data = {}

        self.timestamp = -1
        self.coord_ball = (0, 0, 0)
        self.khep = {}

        self.updated = threading.Event()

    def update_info(self):
        """
        Update the informations sended by the function get_info.
        """
        # Set the timestamp of the calcul

        # Get the Ball's coordinate
        if self.ball is not None:
            # And set the coordinates depending on the team
            coord_ball = (self.ball.x * self.team, -self.ball.z * self.team)

        # Get the Kheperas' coordinates
        new_khep = {}
        for id_khep in self.kheperas.keys():
            new_khep[id_khep] = self.kheperas[id_khep].get_coord()
            # Set the coordinates depending on the team
            add_angle = pi * ((1-self.team)/2)
            new_khep[id_khep] = (new_khep[id_khep][0] * self.team,
                                 new_khep[id_khep][1] * self.team,
                                 (new_khep[id_khep][2] + add_angle) % (2 * pi))
        # The calcul is done, it is flagged as
        self.data = {Coordinate.TIMESTAMP:  time(),
                     Coordinate.BALL     : coord_ball,
                     Coordinate.KHEPERAS : new_khep}
        self.updated.set()


    def get_info(self, timestamp=-1, time_out=None):
        """
        Return informations from the field. If the informations are older than
        the timestamp, we block until the informations are updated (@see
        update_info). A timeout can be set not to wait for ever (by default).

        Be careful, this class is not stand alone besauce of this function.
        update_info should be call in a separate Thread.
        If not, update_info will not be called until the time out is over.

        Return a dict with these key-value pairs :
        timestamp -> float: timestamp in epoch format of object datas.
        ball_coord -> (ball_x:float, ball_y:float): tuple of coordinate of the
            ball (ball_x, ball_y).
        khep_coord -> {id_khepera:int : (x:float, y:float, angle:float)}: dict
        of (int, tuple) pairs:
            id_khepera is the identifier of the khepera
            x, y are coordinates of this khepera
            angle is the angle (in radians) of the orientation of the khepera
        Distance are in meters.
        """
        # If the specified timestamp is equal to or older than the current
        # timestamp, let's update infos
        if timestamp >= self.data[Coordinate.TIMESTAMP]:
            # It's not thread safe here. We can lose an event, and wait for the
            # next one, but it is not a problem here.
            self.updated.clear()
            #   <- we can lose the event here actualy
            self.updated.wait(time_out)

        return self.data


class KheperaSerial(IKheperaSerial):
    """
    Fake Serial port used to communicate with a khepera
    """
    def __init__(self, khepera):
        """
        Initialize attribute of the object.
        """
        self._khepera = khepera

    def read(self, size=1): # pylint: disable-msg=W0613
        """
        Read a message from the fake serial port, if any.
        
        size does not mean anything in our case.
        """
        return None

    def write(self, data):
        """
        Output the given string over the fake serial port.
        
        Decode the commande.
        data: string like this : 'D,LEFT_SPEED,RIGHT_SPEED'
            LEFT_SPEED and RIGHT_SPEED must be numeric, in tick/s.
        """
        if data and (data[0] == Control.MOVE_COMMAND):
            (_, left_speed, right_speed) = data.split(u",")
            khep = self._khepera
            speed_ratio = khep.PULSE_SPEED_RATIO * \
                         (khep.highest_speed / khep.max_speed)

            left_speed  = int( left_speed) * speed_ratio
            right_speed = int(right_speed) * speed_ratio
            self._khepera.new_order(left_speed, right_speed)


class Video(IVideo):
    """Stream of images comming from sonere to rie"""
    def __init__(self, camera):
        """
        Set camera to screenshot.

        Camera: the camera used for screenshot. It must be a Screenshoter
        """
        self.camera = camera
        self.timestamp = -1

    def get_frame(self, timestamp=None, time_out=None):
        """
        Get the most recent image available.
        Block until a new image is found, except if time_out expire
        Return an image newer than timestamp, 
        by default, return a newer image than before
        :return:
            StringIO of a jpeg image
        """
        if not timestamp:
            timestamp = self.timestamp
        img, self.timestamp = self.camera.get_image(timestamp, time_out)
        if img is not None:
            string = StringIO()
            img.save(string, 'jpeg')
            string.seek(0)
            return string
        else:
            return None


class DistantProxy(Pyro.core.ObjBase):
    """
    Pyro proxy for all the interface objects.
    """

    def __init__(self, interface):
        """
        Extends ObjBase.__init__()
        """
        Pyro.core.ObjBase.__init__(self)
        self.__interface = interface

    def __getattr__(self, name):
        return getattr(self.__interface, name)
