#!/usr/bin/env python
# -*- coding:utf-8 -*-
#pylint: disable-msg=E1101,W0231,W0233

"""
This module defines the 3D elements used by the simulator.
"""

import sys
import copy
from math import radians, pi
from itertools import chain

import soya
import soya.sdlconst
import soya.sphere
import soya.cube
import soya.label3d
from soya import Camera, Body, Vector

from Queue import Queue
from scor.sonere import cylinder
import threading
from time import time
from . import interfaces
from .wheel import MotoredAxleJoint
from . import categories


class Khepera(soya.World):
    """
    A Khepera is a World composed by a cylindrical body and two wheels.
    """

    # The timer of the led
    TIMER = 100

    # The ratio tick/s -> m/s
    # 1 tick/s -> PULSE_SPEED_RATIO m/s
    PULSE_SPEED_RATIO = 0.008

    def __init__(self, parent, autobuild=True, x=0, y=0, z=0, #pylint: disable-msg=C0103,C0301
            symbol_material=soya.DEFAULT_MATERIAL):
        """Extend soya.World.__init__. Initialize physical parameters.
        
        x, y, z: initial coordinates of the Khepera in parent.
        autobuild: if True, the model of the Khepera will be built during this
        initialization
        material: the material to be displayed above the Khepera.
        """
        soya.World.__init__(self, parent)
        self.set_xyz(x, y, z) #pylint: disable-msg=E1101
        self.symbol_material = symbol_material

        self.body_radius = 0.035
        self.body_height = 0.03
        self.body_weight = 0.08

        self.wheel_radius = 0.016
        self.wheel_width = 0.0027
        self.wheel_weight = 0.01
        # Distance between axis of the khepera and the wheel
        self.wheel_offcenter = 0.032
        # Distance between axis of th wheel and the bottom face of the body
        self.wheel_alt = 0.0023
        self.holder_radius = 0.005

        # The led dimension
        self.led_radius = 0.005
        # The led timer
        self.led_timer = self.TIMER
        
        # max linear speed of the khepera, in m/s
        self.highest_speed = 1
        self.max_speed = 1
        # max pulse (tick/s) of the khepera
        # 1 tick/s -> 8 mm/s
        self.max_pulse = 127

        self.body        = Body(self)
        self.left_wheel  = Body(self)
        self.right_wheel = Body(self)

        # The holders to stabilize the khepera
        self.front_holder = Body(self)
        self.back_holder  = Body(self)
        self.led1         = Body(self)
        self.body_material = soya.DEFAULT_MATERIAL

        self.left_motor   = None
        self.right_motor  = None

        # Serial to be ThreadSafe
        self._commandes = Queue()
        # serial port (for pyro server)
        self.port = None

        self.pos_init = None

        if autobuild:
            self.build()


    def build(self):
        """
        Call the corrects functions to build a khepera (build_model & build_ode)
        """
        # Save the position in order to reinitate coordinates
        self.pos_init = self.position()
        self.build_model()
        self.build_ode()


    def build_model(self):
        """
        Build the model and the geometry of the Khepera, according to its attribute.
        """
        body = self.body
        body.model = cylinder.Cylinder(radius=self.body_radius,
            material=self.body_material, length=self.body_height).to_model()
        body.z = self.body_height/2

        wheel_material = soya.Material()
        wheel_model = cylinder.Cylinder(radius=self.wheel_radius,
            material=wheel_material, length=self.wheel_width, slices=10)\
            .to_model()
        right_wheel = self.right_wheel
        right_wheel.model = wheel_model
        right_wheel.rotate_x(-90) #pylint: disable-msg=E1101
        right_wheel.z = self.wheel_alt
        right_wheel.y = -self.wheel_offcenter

        left_wheel = self.left_wheel
        left_wheel.model = wheel_model
        left_wheel.rotate_x(90) #pylint: disable-msg=E1101
        left_wheel.z = self.wheel_alt
        left_wheel.y = self.wheel_offcenter

        front_holder = self.front_holder
        front_holder.model = soya.sphere.Sphere(slices=5, stacks=5).to_model()
        front_holder.scale(self.holder_radius, self.holder_radius, #pylint: disable-msg=E1101,C0301
                self.holder_radius)
        front_holder.x = -0.8 * self.body_radius
        front_holder.z = (self.wheel_alt - self.wheel_radius +
                self.holder_radius)

        back_holder = self.back_holder
        back_holder.model = soya.sphere.Sphere(slices=5, stacks=5).to_model()
        back_holder.scale(self.holder_radius, self.holder_radius, #pylint: disable-msg=E1101,C0301
                self.holder_radius)
        back_holder.x = 0.8 * self.body_radius
        back_holder.z = (self.wheel_alt - self.wheel_radius +
                self.holder_radius)

        # Add an emmissive sphere on the top of the khepera for led simulation
        led_material = soya.Material()
        led_material.emissive = (1, 0, 0, 1) # Red
        self.led1.model = soya.sphere.Sphere(material=led_material, slices = 8, stacks =
                4).to_model()
        self.led1.z = self.body_height
        led_radius = self.led_radius
        self.led1.scale(led_radius, led_radius, led_radius)
        # The led is hidden by default. It will be shown with keyboard control
        self.led1.visible = False

        # Rotate the khepera, once it is created.
        self.rotate_x(-90) #pylint: disable-msg=E1101


    def build_ode(self):
        """
        Build the ode_part (GeomBox, mass...) of the Khepera.
        """
        self.body.mass = soya.CylindricalMass(self.body_weight, direction=3,
                r=self.body_radius, h=self.body_height, mode='total_mass')

        self.left_wheel.mass = soya.CylindricalMass(self.wheel_weight,
            direction=3, r=self.wheel_radius, h=self.wheel_width,
                mode='total_mass')

        self.right_wheel.mass = soya.CylindricalMass(self.wheel_weight,
            direction=3, r=self.wheel_radius, h=self.wheel_width,
                mode='total_mass')

        # XXX use Capsule instead of GeomCylinder because Cylinder can't colide each other (ODE limitation)
        geom = soya.GeomCapsule(self.body, self.body_radius, self.body_height)
        if hasattr(geom, 'grip'):
            geom.grip = 0
        else:
            from warnings import warn
            warn("ALERT ALERT SOYA NOT UPGRADED")
        soya.GeomCylinder(self.right_wheel, self.wheel_radius, self.wheel_width)
        soya.GeomCylinder(self.left_wheel, self.wheel_radius, self.wheel_width)

        #self.geom.space = self.space

        axis = Vector(self.left_wheel, 0, 0, 1)
        self.left_motor = MotoredAxleJoint(self.body, self.left_wheel, axis)
        self.left_motor.strength = 0.008


        axis = Vector(self.right_wheel, 0, 0, -1)
        self.right_motor = MotoredAxleJoint(self.body, self.right_wheel, axis)
        self.right_motor.strength = 0.008

        self.front_holder.mass = soya.SphericalMass(0.01,
                                            self.holder_radius, u'total_mass')
        soya.GeomSphere(self.front_holder, self.holder_radius)
        front_holder_axis = soya.BallJoint(self.body, self.front_holder)
        front_holder_axis.anchor = soya.Point(self.front_holder)

        self.back_holder.mass = soya.SphericalMass(0.01,
                                           self.holder_radius, u'total_mass')
        soya.GeomSphere(self.back_holder, self.holder_radius)
        back_holder_axis = soya.BallJoint(self.body, self.back_holder)
        back_holder_axis.anchor = soya.Point(self.back_holder)

        self.led1.mass = soya.SphericalMass(0.01, 0.1, u'total_mass')
        soya.FixedJoint(self.body, self.led1).setFixed()


        self.body.geom.collide_bits = (~0) ^ (categories.WHEEL |
                categories.GROUND)
        self.body.geom.category_bits = categories.ROBOT

        others = (self.left_wheel, self.right_wheel,
                  self.front_holder, self.back_holder)
        for other in others:
            other.geom.collide_bits  = (~0) ^ categories.ROBOT
            other.geom.category_bits = categories.WHEEL

    def begin_round(self):
        """
        Read commands from fake serial input, if any.

        Extends soya.Body.begin_round().
        """
        soya.World.begin_round(self)
        #Read the command type :
        self.read_order()
        # decrement led timer
        if self.led_timer > 0:
            self.led_timer -= 1
            if self.led_timer == 0:
                self.led1.visible = False


    def init_coordinates(self):
        """
        Put the khepera at its orginal coordinate, without any force
        """
        #pylint: disable-msg=W0201
        self.angular_velocity = None #pylint: disable-msg=C0301
        self.linear_velocity  = None #pylint: disable-msg=C0301
        top_vector = (Vector(self.body,0,0,1) % self).z
        if top_vector <= 0:
            self.body.rotate_vertical(180)
        self.add_vector(self.body.position().vector_to(self.pos_init))

    def read_order(self):
        """
        Read commands from the serial input, if any.

        Can throw TypeError or ValueError if the order is incorrect.
        """
        # We apply a command only if there is data in the pipe
        if self._commandes and not self._commandes.empty():
            commande = self._commandes.get()
            if commande is not None:
                #Read the command type :
                (left_speed, right_speed) = commande
                # Command the engines
                self.set_speed(left_speed, right_speed)


    def new_order(self, left_speed, right_speed):
        """
        Send the order 'order' to the Khepera via its serial
        """
        self._commandes.put((left_speed, right_speed))


    def set_speed(self, left_speed, right_speed):
        """
        Set the speed of the Khepera.

        The two wheels will have the same speed.
        left_speed, right_speed: float : speed of the wheel (resp. left or
        right), in percent of max speed. Can be negative. [-1,1]
        """
        # Conversion m/s -> rad/s :
        # speed (rad/s) = speed(m/s) / radius (m)
        self.left_motor.angular_speed = ((left_speed * self.max_speed) /
                self.wheel_radius)
        self.right_motor.angular_speed = ((right_speed * self.max_speed) /
                self.wheel_radius)


    def get_speed(self):
        """
        Get the speed of the Khepera.

        return (left_speed, right_speed): float : speed of the wheel (resp.
        left or right), in percent of max speed. Can be negative. [-1,1] 
        """
        # Conversion rad/s -> m/s :
        # speed (m/s) = speed(rad/s) * radius (m)
        return (self.left_motor.angular_speed * self.wheel_radius /
                self.max_speed,
               self.right_motor.angular_speed * self.wheel_radius /
               self.max_speed)

    def turn(self, angle):
        """
        Rotate the Khepera around its vertical axis, of angle.

        For debugging purpose only : in simulation, the khepera must be
        manipulated only with its motors.

        angle: degrees
        """
        # Rotating around the parent's X axis
        self.body.rotate_z(angle)

    def get_coord(self):
        """
        Return the x, y and angle of the khepera in the parent.
        """
        position = self.body % self.parent
        x = position.x #pylint: disable-msg=C0103
        y = -position.z #pylint: disable-msg=C0103
        parent_ref  = Vector(self.parent, 1, 0, 0)
        khep_forward = Vector(self.body, 1, 0, 0)

        angle = radians(parent_ref.angle_to(khep_forward))

        # Test wether the angle is positive or negative
        if parent_ref.cross_product(khep_forward).y < 0:
            angle *= -1
            angle %= 2 * pi

        return x, y, angle


class Terrain(soya.World):
    """
    A Terrain is a World composed by the ground and the walls.

    It is also composed of the geometry of the terrain.
    A description and a schema of the terrain can be found in doc/
    """

    def __init__(self, parent, autobuild=True, width=7, length=12,
            wall_height=0.5, offside=0.7, border=0.1, middle_wall=0.2,
            wall_material=soya.DEFAULT_MATERIAL,
            offside_material=soya.DEFAULT_MATERIAL,
            border_material=soya.DEFAULT_MATERIAL,
            inner_material=soya.DEFAULT_MATERIAL,
            middle_wall_material=soya.DEFAULT_MATERIAL):
        """Build (or load) the terrain model.

        Extends soya.World.__init__.

        Parameters:
        autobuild: if True, the model of the terrain will be built during this
        initialization. If False, Terrain.build_model() must be called.
        width: terrain width
        length: terrain length
        wall_height: height of the walls
        offside: width of the offside band
        border: width of the line between inner terrain and offside band.
        middle_wall: width of the middle walls.
        {wall,offside,border,inner,middle_wall}_material: material for the
        corresponding element.

        """
        soya.World.__init__(self, parent)
        # setting a lot of parameters
        self.width = float(width)
        self.length = float(length)
        self.wall_height = float(wall_height)
        self.offside = float(offside)
        self.border = float(border)
        self.middle_wall = float(middle_wall)
        self.wall_material = wall_material
        self.offside_material = offside_material
        self.border_material = border_material
        self.inner_material = inner_material
        self.middle_wall_material = middle_wall_material
        #self.filename = "terrain"

        #self._build_or_load_model()
        self.model = None
        if autobuild:
            self.build_model()


    def build_model(self):
        """
        Build the Model of the terrain and save it.
        """

        def innerize(vertex, gap=self.offside):
            "Bring the vertex nearer (of gap) to the center (0,0)."
            # The central line vertex are not brought nearer.
            vertex.x -= abs(vertex.x) >= (self.width / 3) and gap*vertex.x/abs(vertex.x) or 0
            vertex.z -= abs(vertex.z) >= (self.length / 3) and gap*vertex.z/abs(vertex.z) or 0
            return vertex

        def innerize_border(vertex):
            "Bring the vertex nearer (of border) to the center (0,0)."
            return innerize(vertex, self.border)

        # The world used to create the model
        terrain_world = soya.World()
        # The height (4 at the middle) outer corners of the terrain
        corners = [soya.Vertex(terrain_world,
                       -self.width/2, 0, -self.length / 2, 0, 0),
                   soya.Vertex(terrain_world,
                       -self.width/2, 0, -self.border / 2, 1, 0),
                   soya.Vertex(terrain_world,
                       -self.width/2, 0, self.border / 2, 1, 0),
                   soya.Vertex(terrain_world, -self.width/2, 0, self.length/2,
                       0, 0),
                   soya.Vertex(terrain_world, self.width/2, 0, self.length/2, 0,
                       1),
                   soya.Vertex(terrain_world,
                       self.width/2, 0, self.border / 2, 1, 1),
                   soya.Vertex(terrain_world,
                       self.width/2, 0, -self.border / 2, 1, 1),
                   soya.Vertex(terrain_world, self.width/2, 0, -self.length/2,
                       0, 1)]
        # The height inner corners of the terrain
        inner_corners = copy.deepcopy(corners)
        inner_corners = [innerize(corner) for corner in inner_corners]
        border_corners = copy.deepcopy(inner_corners)
        border_corners = [innerize_border(corner) for corner in border_corners]
        # Construct the upper corner of the walls.
        wall_corners = copy.deepcopy(corners)
        for corner in wall_corners:
            corner.y += self.wall_height
        # The ground Faces
        # z positive
        ground_pos = soya.Face(terrain_world, material=self.inner_material)  
        # z negative
        ground_neg = soya.Face(terrain_world, material=self.inner_material)  
        # Create the 8 walls (4 for each length)
        walls = [soya.Face(terrain_world, material=self.wall_material)
                for i in xrange(len(corners))]
        for wall in walls:
            wall.double_sided = True
        # Create the 8 offside faces
        offsides = [soya.Face(terrain_world, material=self.offside_material)
                for i in xrange(len(corners))]
        # Create the 8 border faces
        borders = [soya.Face(terrain_world, material=self.border_material) for i
                in xrange(len(corners))]
        # Add the corners to the ground (offside, border...) and to the walls.
        for corner_num in xrange(len(corners)):
            next_corner_num = (corner_num + 1) % len(corners) 
            corner = corners[corner_num]  # current corner
            next_corner = corners[next_corner_num]  # adjacent corner
            wall_corner = wall_corners[corner_num]
            next_wall_corner = wall_corners[next_corner_num]
            inner_corner = inner_corners[corner_num]
            next_inner_corner = inner_corners[next_corner_num]
            border_corner = border_corners[corner_num]
            next_border_corner = border_corners[next_corner_num]
            # add the two corners, and the upper vertices to the current wall
            walls[corner_num].add(corner)
            walls[corner_num].add(next_corner)
            walls[corner_num].add(next_wall_corner)
            walls[corner_num].add(wall_corner)
            # add the two corners, and the inner corners to the current offside
            # face
            offsides[corner_num].add(corner)
            offsides[corner_num].add(next_corner)
            offsides[corner_num].add(next_inner_corner)
            offsides[corner_num].add(inner_corner)
            # add the two inner corners, and the border corners to the current
            # border face
            borders[corner_num].add(inner_corner)
            borders[corner_num].add(next_inner_corner)
            borders[corner_num].add(next_border_corner)
            borders[corner_num].add(border_corner)

            # Add the corner to the ground (inner faces)
            if corner.z <= 0 and border_corner not in ground_neg.vertices:
                ground_neg.add(border_corner) #pylint: disable-msg=E1101
            if corner.z >= 0 and border_corner not in ground_pos.vertices:
                ground_pos.add(border_corner) #pylint: disable-msg=E1101

        central_line = soya.Face(terrain_world, material=self.border_material)
        central_line.add(border_corners[1])
        central_line.add(border_corners[2])
        central_line.add(border_corners[5])
        central_line.add(border_corners[6])


        # Adjusting the texture for each face of the walls
        for wall in walls:
            for vertex in [vertex for vertex in wall.vertices if vertex.y > 0]:
                vertex.tex_y = self.wall_height

        # Creating two wall at the middle of the terrain
        middle_walls = [soya.cube.Cube(terrain_world,
            material=self.middle_wall_material) for i in xrange(2)]
        for middle_wall in middle_walls:
            middle_wall.y = self.wall_height/2
            middle_wall.scale(self.offside, self.wall_height, self.middle_wall)
        middle_walls[0].x = self.width/2 - self.offside/2
        middle_walls[1].x = -self.width/2 + self.offside/2

        self.model = terrain_world.to_model() #pylint: disable-msg=E1101
        
        self._build_geometry()
        # Save the Model.
        #self.model.filename = self.filename
        #self.model.save()
        north_label = soya.label3d.Label3D(self)
        north_label.set_text("N")
        north_label.add_xyz(0, 2*self.wall_height, -self.length/2)
        north_label.set_size(0.004)
        north_label.set_lit(False)
        south_label = soya.label3d.Label3D(self)
        south_label.set_text("S")
        south_label.add_xyz(0, 2*self.wall_height, self.length/2)
        south_label.set_size(0.004)
        south_label.set_lit(False)
#        north_cube = soya.cube.Cube(self, size=0.04)
#        north_cube.add_xyz(0, 2*self.wall_height, -self.length/2)

    def _build_geometry(self):
        """
        Build the geometry (GeomBox, mass...) of the terrain.
        """
        # A box for the ground
        geom_ground = Body(self)
        geom_ground.y = -0.5
        soya.GeomBox(geom_ground, (self.width, 1, self.length))
        geom_ground.gravity_mode = False
        geom_ground.mass = soya.SphericalMass(sys.maxint)
        geom_ground.pushable = False
        geom_ground.geom.collide_bits  = (~0) ^ categories.ROBOT
        geom_ground.geom.category_bits = categories.GROUND



        # Four boxes for walls, and two in the middle of the terrain
        geom_walls = [Body(self) for i in xrange(6)]
        for geom_wall in geom_walls:
            geom_wall.gravity_mode = False
            geom_wall.mass = soya.SphericalMass(sys.maxint)

        geom_walls[0].x = -self.width/2 - 0.5
        geom_walls[0].pushable = False
        soya.GeomBox(geom_walls[0], (1, self.wall_height, self.length))
        geom_walls[2].x = self.width/2 + 0.5
        geom_walls[2].pushable = False
        soya.GeomBox(geom_walls[2], (1, self.wall_height, self.length))
        geom_walls[1].z = -self.length/2 - 0.5
        geom_walls[1].pushable = False
        soya.GeomBox(geom_walls[1], (self.width, self.wall_height, 1))
        geom_walls[3].z = self.length/2 + 0.5
        geom_walls[3].pushable = False
        soya.GeomBox(geom_walls[3], (self.width, self.wall_height, 1))
        geom_walls[4].x = self.width/2 - self.offside/2
        geom_walls[4].pushable = False
        soya.GeomBox(geom_walls[4], (self.offside, self.wall_height,
            self.middle_wall))
        geom_walls[5].x = -self.width/2 + self.offside/2
        geom_walls[5].pushable = False
        soya.GeomBox(geom_walls[5], (self.offside, self.wall_height,
            self.middle_wall))

        for geom in (wall.geom for wall in geom_walls):
            geom.category_bits = categories.WALL


class Ball(Body):
    """A Ball is a simple sphere used to entertain the Khepera."""

    def __init__(self, parent, x=0, y=0, z=0, #pylint: disable-msg=C0103
            radius=0.3, material=soya.DEFAULT_MATERIAL, autobuild=True):
        """Extend soya.Body.__init__
        
        x: x coordinate in parent.
        y: y coordinate in parent.
        z: z coordinate in parent.
        radius: the radius of the ball.
        material: the material (soya.Material) of the ball.

        """
        Body.__init__(self, parent)
        self.radius = float(radius)
        self.material = material
        # Coulomb coefficient for friction with khepera
        self.mu_khepera = 1
        # % Speed lost while rolling on the Terrain
        self.speed_lost = 0
        # bouciness of the ball on the Terrain and the Khepera
        self.bounciness = 0
        self.model = None
        self.weight = 1
        self.x_init = self.x = x #pylint: disable-msg=C0103
        self.y_init = self.y = y #pylint: disable-msg=C0103
        self.z_init = self.z = z #pylint: disable-msg=C0103
        if autobuild:
            self.build_model()

    def build_model(self):
        """
        Build the model and the geometry of the ball, according to its attribute.
        """
        # Save the position in order to reinitate coordinates
        self.pos_init = self.position()
        self.model = soya.sphere.Sphere(None, material=self.material, slices=10,
                stacks=5).to_model()
        self.scale(self.radius, self.radius, self.radius) #pylint: disable-msg=E1101,C0301

        self.mass = soya.SphericalMass(self.weight, self.radius, 'total_mass')
        if self.geom is None: #pylint: disable-msg=E1101
            soya.GeomSphere(self, self.radius)
        else:
            self.geom.radius = self.radius #pylint: disable-msg=E1101
        self.geom.category_bits = categories.BALL

    def init_coordinates(self):
        """
        Put the ball at its orginal coordinate, without any force
        """
        self.angular_velocity = None #pylint: disable-msg=C0301
        self.linear_velocity  = None #pylint: disable-msg=C0301
        #pos = soya.Position(self.parent, self.y_init, self.z_init)
        #self.add_vector(self.position().vector_to(self.pos_init))
        self.move(self.pos_init)
        #self.set_xyz(self.x_init, self.y_init, self.z_init)

    def hit(self, other, contacts):
        """
        Set the parameters of the contact beetween khepera and ball.
        """
        if isinstance(other.parent, Khepera):
            for contact in contacts:
                contact.bounce = self.bounciness # use dedicated geom attributs
        elif isinstance(other.parent, Terrain):
            for contact in contacts:
                contact.bounce = self.bounciness # use dedicated geom attributs
                contact.motion1 = 10
                self.add_force(self.linear_velocity * - self.speed_lost)


class Controller(object):
    """
    Use to interpret keybord commands
    """

    def __init__(self, terrain, ball, control, coordinate=None, id_khepera=1, 
            kheperas=None):
        """
        Initialize control object.
        """
        self.control = control
        self.ball = ball
        self.terrain = terrain
        self.id_khepera = id_khepera
        self.coordinate = coordinate
        self.age = 0
        self.select_khepera()
        if not kheperas:
            if control:
                kheperas =control._kheperas
            else:
                kheperas = {}
        self.kheperas = kheperas

    def set_id_khepera(self, id_khepera):
        """
        Set the current controlled khepera id.
        """
        self.id_khepera = id_khepera
        self.select_khepera()
    
    def change_khepera(self):
        """
        Set the current khepera to the next
        """
        if not self.control:
            return
        keys = self.kheperas.keys()
        try:
            index = keys.index(self.id_khepera)
        except ValueError:
            index = -1
        self.id_khepera = keys[(index + 1) % len(keys)]
        self.select_khepera()

    def select_khepera(self):
        if self.control:
            # Switch on the light of the controlled khepera
            self.control.switch_light(self.id_khepera, True)


    def init_all_coordinates(self):
        """
        Put the ball and the kheperas at their original coordinates
        """
        self.ball.init_coordinates()
        for khep in self.kheperas.values():
            khep.init_coordinates()

    def process_events(self):
        """
        Catch keyboard events for controling.

        Extends soya.Body.begin_round().
        """

        speed = 1   # khepera's speed
        rotate = 0.7# khepera's speed for rotation

        for event in soya.MAIN_LOOP.events:

            if event[0] == soya.sdlconst.KEYDOWN:
                # Quit Soya
                if event[1] == soya.sdlconst.K_ESCAPE:
                    soya.MAIN_LOOP.stop()
                # Show coordinates
                elif event[1] == soya.sdlconst.K_x:
                    if self.coordinate:
                        print self.coordinate.get_info(time_out=0)
                # Move selected Khepera
                elif event[1] == soya.sdlconst.K_UP and self.control: # move UP
                    self.control.move(self.id_khepera, speed)
                elif event[1] == soya.sdlconst.K_DOWN and self.control: # move DOWN
                    self.control.move(self.id_khepera, -speed)
                elif event[1] == soya.sdlconst.K_LEFT and self.control: # move LEFT
                    self.control.rotate(self.id_khepera, -rotate)
                elif event[1] == soya.sdlconst.K_RIGHT and self.control: # move RIGHT
                    self.control.rotate(self.id_khepera, rotate)
                # Stop selected Khepera
                elif event[1] == soya.sdlconst.K_SPACE and self.control:
                    self.control.stop(self.id_khepera)
                # Select next Khepera
                elif event[1] == soya.sdlconst.K_TAB and self.control:
                    self.control.switch_light(self.id_khepera, False)
                    self.change_khepera()
                    self.control.switch_light(self.id_khepera, True)
                # Select Khepera by id
                elif (event[1]>=soya.sdlconst.K_0 and \
                      event[1]<=soya.sdlconst.K_9):
                    self.set_id_khepera(event[1]-soya.sdlconst.K_0)
                elif event[1] == soya.sdlconst.K_i: # reinit the ball
                    self.ball.init_coordinates()
                elif event[1] == soya.sdlconst.K_o: # reinit the ball and Kheps
                    self.init_all_coordinates()
            elif event[0] == soya.sdlconst.QUIT:
                # Quit Soya
                soya.MAIN_LOOP.stop()
                

class TravelingCamera(soya.Camera):

    def __init__(self, parent, width, length, angle=0.01, delta=0.05):
        soya.Camera.__init__(self, parent)
        # dimension of the field the camera will traveling around
        self.field_width = width
        self.field_length = length
        self.angle = angle     # camera's angle rate
        self.delta = delta     # translated distance of the camera
        self.target = None     # objectif the camera is watching
        self.distance = 1      # need initialization, depending of the init pos
        # booleans used to move or turn the camera
        self.must_turn_up = \
            self.must_turn_down = \
            self.must_turn_left = \
            self.must_turn_right = \
            self.must_translate_left = \
            self.must_translate_right = \
            self.must_move_freely = False

    def look_at(self, target):
        """
        Look at the target, and save it.

        Extends soya.Camera.look_at
        """
        soya.Camera.look_at(self, target)
        self.target = target

    # Rotate the Camera
    #TODO : change x,y,z values when coordinates will be good
    def look_up(self):
        self.move_target(Vector(self,0,0,-1) % self.parent)
    def look_down(self):
        self.move_target(Vector(self,0,0,1) % self.parent)
    def look_left(self):
        self.move_target(Vector(self,-1,0,0) % self.parent)
    def look_right(self):
        self.move_target(Vector(self,1,0,0) % self.parent)

    def move_target(self, vec):
        """
        Move the target of the camera in the plan x0y, in the direction of vec.

        The vector will be normalized so that the distance of the move equals self.angle.
        The z component is put to zero.

        vec: Vector in the coordinate you want to move the target.
        """
        if self.target == None:
            #We don't move the target if it doesn't exist
            return
        #projection in xOy plan (the target only move in this plan)
        #TODO: change vec.y to vec.z when coordinates will be good
        vec.y = 0
        vec.normalize()
        self.target.add_mul_vector(self.angle, vec)
        self.look_at(self.target)
 
    def move_camera_in_plan(self,vec):
        """
        Move the camera in the plan x0y, in the direction of vec.

        The vector will be normalized so that the distance of the move equals
        self.delta.
        The z component is put to zero.

        vec: Vector in the coordinate you want to move the camera.
        """
        #projection in xOy plan (the camera only move in this plan)
        #TODO: change vec.y to vec.z when coordinates will be good
        vec.y = 0
        vec.normalize()
        self.add_mul_vector(self.delta, vec)
        self.look_at(self.target)


    def translate_left(self):
        """
        Translate the camera to the left around the field where the camera is
        looking at.
        """
        self._translate(1)

    def translate_right(self):
        """
        Translate the camera to the right around the field where the camera is
        looking at.
        """
        self._translate(-1)

    def _translate(self,signe=1): # signe>0:left, signe<0:right
        """
        Translate the camera to the left or to the right.

        signe:  
            if signe>=0 : translate to the left
            if signe <0 : translate to the right
        """
        #TODO: put the signe '-' off when good coordinates are used
        signe = -signe < 0 and -1 or 1
        delta = float(self.delta)     # the real distance to move
        # We have to always be around the field, 
        # the trajectory is an octogone (@see the doc)
        xmax = self.field_width /2
        ymax = self.field_length/2
        pos = self.position()
        (x,y,z) = pos.x, pos.z, pos.y
        (next_x, next_y, next_z) = x,y,z

        
        #Calcul the distance from the field's border
        (X, Y) = abs(x), abs(y)
        if (Y<=ymax and X<=xmax):
            if (self.distance > 0.01):
                distance = 0.01
            else:
                distance = self.distance
            self.distance = distance
            if (xmax-X <= ymax-Y):
                #we are nearest the xmax border than the ymax border
                ymax = ymax - (xmax-X) - distance
                xmax = X - distance
            else:
                xmax = xmax - (ymax-Y) -distance
                ymax = Y - distance
        if(Y>ymax and X>xmax):
            distance = self.distance
        elif(X>xmax):
            distance = X-xmax
        else:
            distance = Y-ymax
        self.distance = distance

        self.msg = ""
        if (Y>=ymax and X>=xmax):
            # We are at a corner, outside the field
            next_x = self._translate_x(signe, x, y,
                    xmax+distance, ymax, delta/2)
            next_y = self._translate_y(signe, x, y,
                    xmax ,ymax+distance, delta/2)
        if (Y<=ymax and (next_x,next_y)==(x,y)): 
            # parallele or at the limit of the corner
            next_y = self._translate_y(signe, x, y, xmax, ymax, delta)
        if (X<=xmax and (next_x,next_y)==(x,y)): 
            # parallele or at the limit of the corner
            next_x = self._translate_x(signe, x, y, xmax, ymax, delta)
        self.set_xyz(next_x, next_z, next_y)  
        if self.target==None:
            self.target = soya.Point(self.parent, 0, 0, 0)
        soya.Camera.look_at(self, self.target)


    def _translate_y(self, signe, x, y, xmax, ymax, delta):
        """
        Return the next Y position of the caméra

        signe: direction of the translation, @see _translate()
        x,y:  position of the camera
        xmax, ymax: limit of the translation in this round
        delta: distance to move
        """
        if (x>=xmax):
            signe = -signe
        dy = signe * delta
        #TODO: ligne pour debuggage, à supprimer
        self.msg += "\n    translate_y : y=%s ; dy=%s ; ymax=%s" % (y,dy,ymax)
        if (abs(y + dy) >= ymax):
            return y+dy>0 and ymax or -ymax
        else:
            return y+dy

    def _translate_x(self, signe, x, y, xmax, ymax, delta):
        """
        Return the next X position of the caméra

        @see _translate_y()
        """
        if (y<=-ymax):
            signe = -signe
        dx = signe * delta
        #TODO: ligne pour debuggage, à supprimer
        self.msg += "\n    translate_x : x=%s ; dx=%s ; xmax=%s" % (x,dx,xmax)
        if (abs(x + dx) >= xmax):
            return x+dx>0 and xmax or -xmax
        else:
            return x+dx


    def begin_round(self):
        """
        Catch keyboard events for controling.

        Extends soya.Camera.begin_round().
        """

        soya.Camera.begin_round(self)

        for event in soya.MAIN_LOOP.events:

            if event[0] == soya.sdlconst.KEYDOWN:
                # Start moving the Camera
                if event[1] == soya.sdlconst.K_z: # turn UP
                    self.must_turn_up = True
                elif event[1] == soya.sdlconst.K_s: # turn DOWN
                    self.must_turn_down = True
                elif event[1] == soya.sdlconst.K_q: # turn LEFT
                    self.must_turn_left = True
                elif event[1] == soya.sdlconst.K_d: # turn RIGHT
                    self.must_turn_right = True
                elif event[1] == soya.sdlconst.K_a: # translate LEFT
                    self.must_translate_left = True
                elif event[1] == soya.sdlconst.K_e: # translate RIGHT
                    self.must_translate_right = True
                elif event[1] == soya.sdlconst.K_LALT: # move freely the camera
                    self.must_move_freely = True
                elif event[1] == soya.sdlconst.K_c: # want to show the target
                    self.target.visible = not self.target.visible
            elif event[0] == soya.sdlconst.KEYUP:
                # Stop moving the Camera
                if event[1] == soya.sdlconst.K_z: # turn UP
                    self.must_turn_up = False
                elif event[1] == soya.sdlconst.K_s: # turn DOWN
                    self.must_turn_down = False
                elif event[1] == soya.sdlconst.K_q: # turn LEFT
                    self.must_turn_left = False
                elif event[1] == soya.sdlconst.K_d: # turn RIGHT
                    self.must_turn_right = False
                elif event[1] == soya.sdlconst.K_a: # translate LEFT
                    self.must_translate_left = False
                elif event[1] == soya.sdlconst.K_e: # translate RIGHT
                    self.must_translate_right = False
                elif event[1] == soya.sdlconst.K_LALT: # don't move freely cam
                    self.must_move_freely = False
            elif event[0] == soya.sdlconst.QUIT:
                # Quit Soya
                soya.MAIN_LOOP.stop()

        # Move the Camera
        if self.must_move_freely:
            #we only move the camera
            if self.must_turn_up:
                self.move_camera_in_plan(Vector(self,0,0,-1) % self.parent)
            if self.must_turn_down:
                self.move_camera_in_plan(Vector(self,0,0, 1) % self.parent)
            if self.must_turn_left:
                self.move_camera_in_plan(Vector(self,-1,0,0) % self.parent)
            if self.must_turn_right:
                self.move_camera_in_plan(Vector(self, 1,0,0) % self.parent)
            if self.must_translate_left:
                self.add_xyz(0,-self.delta, 0)
                self.look_at(self.target)
            if self.must_translate_right:
                self.add_xyz(0, self.delta, 0)
                self.look_at(self.target)
        else:
            #we only  move the target
            if self.must_turn_up:
                self.look_up()
            if self.must_turn_down:
                self.look_down()
            if self.must_turn_left:
                self.look_left()
            if self.must_turn_right:
                self.look_right()
            if self.must_translate_left:
                self.translate_left()
            if self.must_translate_right:
                self.translate_right()


class Screenshoter(soya.Camera):
    """
    The Camera used to take the screenshot for video simulation.
    """

    def __init__(self, parent):
        """
        Initializes camera_display and img attribute.

        Set counter for regulating screenshot frequency.
        Extends soya.Camera.__init__
        """
        soya.Camera.__init__(self, parent)
        self.display_camera = None
        self.img = None
        self.counter = 1
        self.period = 1
        self.data = None, -1
        self.updated = threading.Event()

    def update_screenshot(self):
        """
        Perform a screenshot every period.
        """
        self.counter += 1
        # Take a screenshot only every period.
        if self.counter > self.period:
            # Reset counter
            self.counter = 1
            self.screenshot()

    def screenshot(self):
        """
        Change cameras, and take a screenshot.
        """
        #Change the root_display of soya for screenshot, and keep the old one to
        #restore it later.
        self.display_camera = soya.root_widget
        soya.set_root_widget(self)
        # Do the rendering and screenshot the backbuffer
        soya.render(swap_buffer=0)
        img = soya.screenshot(use_back_buffer=True)
        self.data = img, time()
        # Set current camera to the normal one
        soya.set_root_widget(self.display_camera)
        self.updated.set()

    def get_image(self, timestamp=-1, time_out=None):
        """
        Return current image from the camera. If the image are older than
        the timestamp, we block until the image is updated
        A timeout can be set not to wait for ever (by default).

        Be careful, this class is not stand alone besauce of this function.

        Return a list with these value:
        timestamp -> float: timestamp in epoch format of object datas.
        image -> PIL.image
        """
        # If the specified timestamp is equal to or older than the current
        # timestamp, let's update infos
        if timestamp >= self.data[1]:
            # It's not thread safe here. We can lose an event, and wait for the
            # next one, but it is not a problem here.
            self.updated.clear()
            #   <- we can lose the event here actualy
            self.updated.wait(time_out)

        return self.data
