#!/usr/bin/env python
# -*- coding:utf-8 -*-
# pylint: disable-msg=R0902,E1101
"""
This module contains the only class Simulator.

A simulator can be started by executing this module, which create and
launch a simulator.
"""

import os
from os.path import dirname
from threading import Thread
from math import tan, radians

import soya
import Pyro.core

from scor.sonere.drag_drop_world import DragDropWorld
from scor.sonere.models import Terrain, Khepera, Ball, Controller
from scor.sonere.models import TravelingCamera, Screenshoter
from scor.sonere.interfaces import Control, Coordinate, KheperaSerial, Video, DistantProxy
from scor.common.config import get_scor_config, ROBOT_SECTION_MASK


def debug(scene, round_number=40):
    """
    Test-purpose function, to view rendering frame-by-frame.
    """
    for _ in range(round_number):
        main_loop = soya.MainLoop(scene)
        main_loop.begin_round()
        main_loop.advance_time(1)
        main_loop.end_round()
        soya.render()


class Simulator(object):
    """The simulator allows to simulate a SCOR game.

    It renders a 3D scene of a terrain, Kheperas and a Ball. Kheperas can be
    controlled and a camera can output images for shape detection.

    A simulator is a Thread : once initialized (created), it must be started:

    >>> sim = Simulator()
    >>> sim.start()

    """

    _KHEPERAS_SECTION     = u"robots"
    _TERRAIN_SECTION      = u"terrain"
    _BALL_SECTION         = u"ball"
    _SCREENSHOTER_SECTION = u"screenshoter"

    VIDEO      = u"video"
    COORDINATE = u"coordinate"
    CONTROL    = u"control"
    SERIAL     = u"serial"


    def __init__(self, served = None):
        """
        Load configuration files and create the scene.
        """
        self.config, argv = self._load_config()
        self.config.set_default_teams()
        if served:
            self.served = served
        else:
            self.served = self._parse_argv(argv)

        # adding our data directory
        #soya.path.append(os.path.basename(__file__ + "data"))
        soya.path.append(os.path.join(os.path.dirname(__file__), "data"))
        self.scene = DragDropWorld()  # Create the scene
        # Create the terrain
        self.terrain = self._config_terrain()
        self.scene.terrain = self.terrain
        # Create the Khepera list
        self.kheperas, self.serials = self._config_kheperas()
        # Create the ball
        self.ball = self._config_ball()
        # Create the cameras who takes the screenshot
        self.screenshoter = None
        self.screenshoter_south = None
        self.screenshoter_north = None
        self._config_screenshoter()

        # Set simulation parameters
        self.main_loop = soya.MainLoop(self.scene)
        self.main_loop.round_duration = 0.014


        # The control interface
        if Simulator.CONTROL in self.served:
            self.control = Control(self.kheperas)
        else:
            self.control = None

        if Simulator.COORDINATE in self.served:
            # The coordinate interface for north team
            self.coordinate_north = Coordinate(self.terrain, self.ball,
                    self.kheperas, Coordinate.NORTH)
            # Call update_info() at each round, for each team
            self.main_loop.round_tasks.append(self.coordinate_north.update_info)

            # The coordinate interface for south team
            self.coordinate_south = Coordinate(self.terrain, self.ball,
                    self.kheperas, Coordinate.SOUTH)
            self.main_loop.round_tasks.append(self.coordinate_south.update_info)
        else:
            self.coordinate_south = None

        if Simulator.VIDEO in self.served:
            # The video interface
            self.video = Video(self.screenshoter)
            # Call update_screenshot() at each round 
            self.main_loop.round_tasks.append(
                    self.screenshoter.update_screenshot)

            # The video interface for south theam
            self.video_south = Video(self.screenshoter_south)
            self.main_loop.round_tasks.append(
                    self.screenshoter_south.update_screenshot)

            # The video interface for the north team
            self.video_north = Video(self.screenshoter_north)
            self.main_loop.round_tasks.append(
                    self.screenshoter_north.update_screenshot)

        self.scene.gravity = soya.Vector(self.scene, 0, -9.8, 0)
        self.scene.cfm *= 10
        # Create a light
        light = soya.Light(self.scene)
        light.set_xyz(0, 5, 0) #pylint: disable-msg=E1101
        # Create and place a camera
        camera = TravelingCamera(self.scene, self.terrain.width,
                self.terrain.length)
        camera.set_xyz(-0.9, 0.7, 0.6) #pylint: disable-msg=E1101
#        camera.set_xyz(-0.1, 0.5, 0.2) #pylint: disable-msg=E1101
        # The objectif of the camera is an invisible cube
        # you can show it by pressing 'c' when simulating the world
        objectif = soya.cube.Cube(self.scene, size=0.03)
        objectif.visible = False
        #objectif = soya.Point(self.scene, 0.1, 0.4, 0.1)
        camera.look_at(objectif)
        self.scene.set_camera(camera)
        soya.set_root_widget(camera)
        

        # The keybord controller
        self.controller = Controller(self.terrain, self.ball,
                self.control, self.coordinate_south, kheperas = self.kheperas)
        # Process keyboard controller at each round
        self.main_loop.round_tasks.append(self.controller.process_events)
        # Process mouse controller at each round
        self.main_loop.round_tasks.append(self.scene.process_events)

    def start(self):
        """
        Initialise soya and launch soya's main loop.
        """
        try:
            #Start soya main loop
            self.main_loop.main_loop()
        finally:
            # When done, quit soya
            soya.quit()

    def _load_config(self, path=None):
        """
        Load the config files : a required default one, and an optionnal custom
        one.
        """
        # Find config files
        if not path == None:
            default_path = path
        else:
            default_path = os.path.join(dirname(__file__), u'data', u'config',
                u"default.ini")
        # Load the config
        config_files = (default_path,)
        return get_scor_config(extra_config_before = config_files)

    def _parse_argv(self, argv):
        """
        Parse arg to see which modules run
        """
        if argv and not 'all' in argv:
            return set(argv)
        else:
            return set(['serial', 'control', 'coordinate', 'video'])

    def _config_element(self, element, section):
        """
        Set element's parameters from section in config file.

        The options (of the specified section) from the config file are used to
        set corresponding attributes of the element.

        element: object to be configured (Terrain, Ball...).
        section: section name from the config file which will be used to
        configure the element.
        """
        image_suffix = u"_image"
        color_suffix = u"_color"
        # Automatically find options with the name of an attribute of the
        # terrain.
        if not self.config.has_section(section):
            return
        for option in self.config.options(section):
            # options of type float
            if hasattr(element, option) and \
                    not callable(getattr(element, option)):
                try:
                    setattr(element, option, self.config.getfloat(section,
                        option))
                except ValueError:
                    # options of type string
                    setattr(element, option, self.config.get(section, option))
            # material (image) options
            if option.endswith(image_suffix):
                # Keep only the attribute name
                attr_name = option[:-len(image_suffix)]
                if hasattr(element, attr_name) and \
                not callable(getattr(element, attr_name)):
                    # Build the material if necessary
                    if getattr(element, attr_name) == soya.DEFAULT_MATERIAL:
                        setattr(element, attr_name, soya.Material())
                    getattr(element, attr_name).texture = \
                    soya.Image.get(self.config.get(section,option))
            # material (color) options
            if option.endswith(color_suffix):
                # Keep only the attribute name
                attr_name = option[:-len(color_suffix)]
                if hasattr(element, attr_name) and \
                not callable(getattr(element, attr_name)):
                    # Build the material if necessary
                    if getattr(element, attr_name) == soya.DEFAULT_MATERIAL:
                        setattr(element, attr_name, soya.Material())
                    getattr(element, attr_name).diffuse = \
                        self.config.getcolor(section,option)

    def _config_terrain(self):
        """
        Create the terrain and set its parameters from config file.
        """
        terrain = Terrain(self.scene, autobuild=False)
        self._config_element(terrain, Simulator._TERRAIN_SECTION)
        terrain.build_model()
        return terrain

    def _config_ball(self):
        """
        Create and configure the ball from the config file.
        """
        ball = Ball(self.scene, autobuild=False)
        self._config_element(ball, Simulator._BALL_SECTION)
        ball.build_model()
        return ball

    def _config_kheperas(self):
        """
        Create the kheperas and set their parameters from config file.
        """
        kheperas = {}
        serials = {}
        for khep_id in self.config.robots_ids():
            khepera = Khepera(self.scene, autobuild=False)
            # Config the dimension of the khepera, which is the same for all
            self._config_element(khepera, Simulator._KHEPERAS_SECTION)
            # Config the position, which depend of the khepera
            self._config_element(khepera, ROBOT_SECTION_MASK % khep_id)
            # Build the khepera
            khepera.build()

            if Simulator.SERIAL in self.served:
                # Create the khepera serial, and connect it to the pyro server
                serials[khepera.port] = KheperaSerial(khepera)

            kheperas[khep_id] = khepera
        return kheperas, serials

    def _config_screenshoter(self):
        """
        Create and configure the screenshoters from the config file.

        Place the cameras at the correct places to display the whole terrain in
        the rendered picture.
        """
        # Configure central camera
        self.screenshoter = Screenshoter(self.scene)
        self._config_element(self.screenshoter, Simulator._SCREENSHOTER_SECTION)
        self.screenshoter.y = self.terrain.width * 1.2 / (2 *
                tan(radians(self.screenshoter.fov / 2)))
        # Look at the center of the scene
        self.screenshoter.look_at(soya.Point(self.scene, 0, 0, 0))
        self.screenshoter.turn_z(-90)

        # Configure south camera
        self.screenshoter_south = Screenshoter(self.scene)
        self._config_element(self.screenshoter_south,
                Simulator._SCREENSHOTER_SECTION + u'_south')
        self.screenshoter_south.z = self.terrain.length / 2
        self.screenshoter_south.y = self.terrain.width * 1.3 / (2 *
                tan(radians(self.screenshoter.fov / 2)))
        self.screenshoter_south.look_at(soya.Point(self.scene, 0, 0, 0.1 *
            self.terrain.length))
        self.screenshoter_south.turn_z(-90)

        # Configure north camera
        self.screenshoter_north = Screenshoter(self.scene)
        self._config_element(self.screenshoter_north,
                Simulator._SCREENSHOTER_SECTION + u'_north')
        self.screenshoter_north.z = -self.terrain.length / 2
        self.screenshoter_north.y = self.terrain.width * 1.3 / (2 *
                tan(radians(self.screenshoter.fov / 2)))
        self.screenshoter_north.look_at(soya.Point(self.scene, 0, 0, -0.1 *
            self.terrain.length))
        self.screenshoter_north.turn_z(-90)


class DistantSimulator(object):
    """
    The distant simulator allow the use of the simulator with distant access
    """

    def __init__(self, sim = None, served = None):
        """
        Construc the distant simulator with the simulator sim
        """
        if not sim:
            self.sim = Simulator(served = served)
        self.scene = self.sim.scene
        self.config = self.sim.config
        self.kheperas = self.sim.kheperas
        self.ball = self.sim.ball
        self.served = self.sim.served
        # Init Pyro server
        self.pyro_servers = {}
        self.pyro_init()
        self.publish_pyro()
        self.server_threads = None

    def pyro_init(self):
        """
        Initialize the pyro server.

        Control, Coordinate and KheperaSerial objects will be connected during
        the creation of them.
        """
        Pyro.core.initServer()
        for served in self.served:
            if self.config.has_option("pyro_server", "%s_port" % served):
                port = int(self.config.get("pyro_server", "%s_port" % served))
                print 'port', port, 'is serving', served
                self.pyro_servers[served] = Pyro.core.Daemon(port=port)

    def publish_pyro(self):
        """
        Publish all wanted modules in pyro server

        Modules are listed in served.
        """
        # The control interface
        if Simulator.CONTROL in self.served:
            self.control = DistantProxy(self.sim.control)
            uri = u'control'
            self.pyro_servers[Simulator.CONTROL].connect(self.control, uri)

        if Simulator.COORDINATE in self.served:
            self.coordinate_north = DistantProxy(self.sim.coordinate_north)
            # The coordinate interface for north team
            uri = u'coordinate/north'
            self.pyro_servers[Simulator.COORDINATE].connect(
                    self.coordinate_north, uri)

            # The coordinate interface for south team
            self.coordinate_south = DistantProxy(self.sim.coordinate_south)
            uri = u'coordinate/south'
            self.pyro_servers[Simulator.COORDINATE].connect(
                    self.coordinate_south, uri)

        if Simulator.VIDEO in self.served:
            self.video = DistantProxy(self.sim.video)
            # The video interface
            uri = u'video/top'
            self.pyro_servers[Simulator.VIDEO].connect(self.video, uri)

            # The video interface for south theam
            self.video_south = DistantProxy(self.sim.video_south)
            uri = u'video/south'
            self.pyro_servers[Simulator.VIDEO].connect(
                    self.video_south, uri)

            # The video interface for the north team
            self.video_north = DistantProxy(self.sim.video_north)
            uri = u'video/north'
            self.pyro_servers[Simulator.VIDEO].connect(
                    self.video_north, uri)

        if Simulator.SERIAL in self.served:
            # Connect khepera serials to the pyro server
            for port, serial in self.sim.serials.items():
                uri = "serial/" + port
                self.pyro_servers['serial'].connect(DistantProxy(serial), uri)


    def start_pyro(self):
        """
        Start pyro in new Thread
        """
        # Start pyro server
        self.server_threads  = [Thread(target=server.requestLoop,
                                       name="pyro_main_%s" % name)
                                for name,server in self.pyro_servers.items()]
        for thread in self.server_threads:
            thread.start()

    def stop_pyro(self):
        """
        Stop the Thread of pyro
        """
        for server in self.pyro_servers.values():
            server.shutdown(True)

    def start(self):
        """
        Initialise soya and launch soya's main loop with pyro.
        """
        try:
            #Start pyro server
            self.start_pyro()
            #Start soya main loop
            self.sim.start()
        finally:
            #Stop pyro server
            self.stop_pyro()

