#!/usr/bin/env python
# -*- coding:utf-8 -*-

from unittest import TestCase, main as unittest_main
import sys
import os

from scor.sonere import simulator, siteconfig

class Simulator_config_TestCase(TestCase):
    """
    Tests on the loading of a config file.

    This TestCase is intended for all config tests of the simulator : to test
    new values, just append them to self.params in setUp(), and add a new test_
    method.
    """

    def setUp(self):
        """
        Write an appropriated custom.cfg config file.

        If custom.cfg already exists, save it.
        """
        self.config = os.path.join(siteconfig.DATADIR, u"custom.ini")
        self.backup = self.config + ".bak"
        if os.path.exists(self.config):
            os.rename(self.config, self.backup)

        self.config_path = "--config="+self.config
        sys.argv.append(self.config_path)

        # Setting the parameters of the config file.
        # Keys of self.params are sections names, values are dict, whose keys are
        # config keys, and values are config value.
        self.params = {}
        
        self.params[u"terrain"] = {
            u"width" : u"17",
            u"length" : u"22",
            u"wall_height" : u"3",
            u"offside" : u"2",
            u"border" : u"1",
            u"middle_wall" : u"2",
            u"wall_material_image" : u"burlwood2.png",
            u"offside_material_image" : u"butcherblock.png",
            u"border_material_color" : u"(0,1,0,1)",
            u"inner_material_image" : u"butcherblock.png",
            u"middle_wall_material_image" : u"burlwood2.png"}

        self.params[u"ball"] = {
            u"x" : "0",
            u"y" : "5",
            u"z" : "-3",
            u"radius" : "0.5",
            u"material_image" : "shirtblack.png"}

        f = open(self.config, u"w")
        for section, options in self.params.items():
            f.write(u"[" + section + u"]\n")
            for key, value in options.items():
                f.write(key + u"=" + value + u"\n")

        f.close()

        self.simulator = simulator.Simulator()
        #self.simulator.run()

    def tearDown(self):
        """
        Delete the custom.cfg config file, and restore the old one.
        """

        del self.simulator

        sys.argv.remove(self.config_path)
        os.remove(self.config)
        if os.path.exists(self.backup):
            os.rename(self.backup, self.config)

    def test_terrain_config(self):
        """
        The simulator must build a terrain according to the parameters.
        """
        self._check_params(self.simulator.terrain, u"terrain")

    def test_ball_config(self):
        """
        The simulator must build a ball according to the parameters.
        """
        self._check_params(self.simulator.ball, u"ball")

    def _check_params(self, object, section):
        """
        Check the value of parameters of the objects.

        It must be equal to those defined in the config file.
        """
        for key, value in self.params[section].items():
            image_suffix = u"_image"
            color_suffix = u"_color"
            if not(key.endswith(image_suffix) or key.endswith(color_suffix)):
                value = float(value)
                self.assertEqual(getattr(object,key), value)
            else:
                # if non-float value
                if key.endswith(image_suffix):
                    filename = getattr(object,
                            key[:-len(image_suffix)]).texture.get_filename()
                    self.assertEqual(filename, value)
                elif key.endswith(color_suffix):
                    rgba = getattr(object,
                            key[:-len(color_suffix)]).diffuse
                    string_tuple = value[1:len(value)-1].split(u",")
                    expected_rgba = tuple([float(i) for i in string_tuple])
                    self.assertEqual(rgba, expected_rgba)


if __name__ == '__main__':
    unittest_main()
