#!/usr/bin/env python
# -*- coding:utf-8 -*-

from unittest import TestCase, main as unittest_main
import sys
import os
from math import degrees
import time

import soya
import Pyro.core
from StringIO import StringIO
import Image

from scor.sonere.simulator import Simulator
from scor.sonere.models import Khepera, Ball, Screenshoter
from scor.sonere.interfaces import Coordinate, Control, KheperaSerial, Video


class Coordinate_TestCase(TestCase):
    """
    Test the Coordinate interface of the simulator.
    """

    DEFAULT_TIME_OUT = None

    def setUp(self):
        self.scene = soya.World()

    def tearDown(self):
        del self.scene

    def test_coordinates(self):
        """
        Test the results provided by Coordinate.
        """
        khep_x = 0.2
        khep_y = 0.3
        khep_z = 0.5
        khep_angle = 350 # degrees for soya
        khep = Khepera(self.scene, x=khep_x, y=khep_y, z=khep_z)
        # Rotating
        khep.turn(khep_angle)
        terrain_ref  = soya.Vector(self.scene, 1, 0, 0)
        khep_forward = soya.Vector(khep.body, 1, 0, 0)

        angle = khep_forward.angle_to(terrain_ref)
        ball_x = 0.3
        ball_y = 0.3
        ball_z = 0.5
        ball = Ball(self.scene, x=ball_x, y=ball_y, z=ball_z)

        coord = Coordinate(self.scene, ball, {0: khep}, Coordinate.SOUTH)

        coord.update_info()
        info = coord.get_info(time_out=Coordinate_TestCase.DEFAULT_TIME_OUT)
        coord_x = info[u'khep_coord'][0][0] # x of khep #1
        coord_y = info[u'khep_coord'][0][1] # y of khep #1
        coord_angle = info[u'khep_coord'][0][2] # angle
        self.assertAlmostEqual(khep_x, coord_x)
        self.assertAlmostEqual(-khep_z, coord_y)
        self.assertAlmostEqual(khep_angle, degrees(coord_angle), 3)

        coord_x = info[u'ball_coord'][0] # x of ball
        coord_y = info[u'ball_coord'][1] # y of ball
        self.assertAlmostEqual(ball_x, coord_x)
        self.assertAlmostEqual(-ball_z, coord_y)

    def test_timestamp_old(self):
        """
        Test that old timestamp makes Coordinate return new info.
        """
        khep = Khepera(self.scene)
        ball = Ball(self.scene)

        coord = Coordinate(self.scene, ball, {0: khep}, Coordinate.SOUTH)

        # Get first info
        coord.update_info()
        info_old = coord.get_info(time_out=Coordinate_TestCase.DEFAULT_TIME_OUT)
        time_old = info_old[u"timestamp"]

        # Change objects
        khep.x = 0.2
        khep.y = 0.3
        khep.z = 0.5
        ball.x = 0.3
        ball.y = 0.3
        ball.z = 0.5

        # Ensure the timestamp will be different
        time.sleep(0.01)
        coord.timestamp = time.time()

        # Get second info
        coord.update_info()
        info_new = coord.get_info(time_old,
                time_out=Coordinate_TestCase.DEFAULT_TIME_OUT)

        self.assertTrue(time_old < info_new[u"timestamp"], "old: %d, new: %d" %
                (time_old, info_new[u"timestamp"]))
        self.failIfEqual(info_new, info_old)

    def test_timestamp_recent(self):
        """
        Test that the same timestamp makes Coordinate return same info.
        (without update)
        """
        TIME_OUT = 1
        khep = Khepera(self.scene)
        ball = Ball(self.scene)

        coord = Coordinate(self.scene, ball, {0: khep}, Coordinate.SOUTH)

        # Get first info
        coord.update_info()
        info_old = coord.get_info(time_out=Coordinate_TestCase.DEFAULT_TIME_OUT)
        time_old = info_old[u"timestamp"]

        # Get second info
        info_new = coord.get_info(time_old, time_out=TIME_OUT)

        self.assertTrue(time_old+TIME_OUT <= time.time())
        self.assertEqual(time_old, info_new[u"timestamp"])
        self.assertEqual(info_new, info_old)


class Control_TestCase(TestCase):
    """
    Test Control interface.

    Serial is covered too, but not formally tested.
    """
    def setUp(self):
        self.scene = soya.World()

    def tearDown(self):
        del self.scene

    def spend_time(self, time):
        self.main_loop = soya.MainLoop(self.scene)
        for i in xrange(time):
            self.main_loop.begin_round()
            self.main_loop.advance_time(1)
            self.main_loop.end_round()

    def test_control(self):
        """
        Test Control calls.
        """
        khep = Khepera(self.scene)
        control = Control({0: khep})
        
        speed = 0.3
        control.move(0, speed)
        # make the begin_round of khepera to be run
        self.spend_time(1)
        self.assertEqual(speed, khep.left_motor._speed * khep.wheel_radius /
                khep.max_speed)
        self.assertEqual(speed, khep.right_motor._speed * khep.wheel_radius /
                khep.max_speed)

        control.rotate(0, speed)
        self.spend_time(1)
        self.assertEqual(speed, khep.left_motor._speed * khep.wheel_radius /
                khep.max_speed)
        self.assertEqual(-speed, khep.right_motor._speed * khep.wheel_radius /
                khep.max_speed)


class KheperaSerial_TestCase(TestCase):
    """
    Test the Class which decode a real Khepera order to a simulated Khepera
    """
    def setUp(self):
        """
        Create the scene.
        """
        self.scene = soya.World()

    def tearDown(self):
        """
        Delete the scene.
        """
        del self.scene

    def test_new_order(self):
        """
        Test the treatment of a new order
        """
        left_speed = 4
        right_speed = -6
        commande = "D,"+str(left_speed)+","+str(right_speed)

        khepera = Khepera(self.scene)
        serial = KheperaSerial(khepera)

        # Send the order
        serial.write(commande)
        
        # Treat the order
        khepera.read_order()
        (l, r) = khepera.get_speed()

        self.assertEqual(l,float(left_speed) * khepera.PULSE_SPEED_RATIO)
        self.assertEqual(r,float(right_speed) * khepera.PULSE_SPEED_RATIO)

    def test_bad_order(self):
        """
        Test the treatment of a new order
        """
        left_speed = 4
        right_speed = -6
        commande = "R,"+str(left_speed)+","+str(right_speed)

        khepera = Khepera(self.scene)
        serial = KheperaSerial(khepera)

        # Send the order
        serial.write(commande)
        
        # Treat the order
        khepera.read_order()
        (l, r)= khepera.get_speed()

        self.assertNotEqual(l,left_speed)
        self.assertNotEqual(r,right_speed)

    

class Video_TestCase(TestCase):
    """
    Test Video interface.
    """

    def setUp(self):
        self.scene = soya.World()
        self.camera = Screenshoter(self.scene)
        soya.set_root_widget(self.camera)

    def tearDown(self):
        del self.scene, self.camera
        soya.quit()

    def test_StringIOImage(self):
        """
        Test the return of get_frame to be a StringIO of a jpeg file.
        """
        soya.init(quiet=True)
        video = Video(self.camera)

        frame = video.get_frame(time_out=0)
        self.assertEqual(frame, None)

        # Create an image
        soya.render()
        self.camera.screenshot()

        frame = video.get_frame(time_out=0)
        self.assertTrue(isinstance(frame, StringIO))

        PILimage = Image.open(frame)
        self.assertEqual(PILimage.format, u'JPEG')


if __name__ == '__main__':
    unittest_main()
