#!/usr/bin/env python
# -*- coding:utf-8 -*-

from unittest import TestCase, main as unittest_main
import sys
import os
import time
import thread
import copy
import math #for math.ceil in TravelingCamera

import soya
from soya.testlib import SoyaTestCase

from scor.sonere import models, cylinder, simulator


class Terrain_geometry_TestCase(TestCase):
    """
    Tests on the correct geometry of the Terrain.

    Testing with balls : where are the balls when they are thrown upon,
    aside... the terrain ?
    """
    def setUp(self):
        """
        Initilize soya, and create the tests objects
        """
        # Initilazing soya
        import soya
        global soya
        soya.path.append(os.path.join(os.path.dirname(sys.argv[0]), "../data"))
        self.scene = soya.World()  # Create the scene
        self.scene.gravity = soya.Vector(self.scene, 0, -9.8, 0)
        # Create a light
        light = soya.Light(self.scene)
        light.set_xyz(10,10,10)
        # Create and place a camera
        camera = soya.Camera(self.scene)
        camera.set_xyz(-5,6,8)
        #camera.set_xyz(0,10,0)
        #camera.look_at(soya.Point(self.scene,0,0,0))
        camera.look_at(soya.Point(self.scene,0,0,2))
        soya.set_root_widget(camera)
        self.terrain = models.Terrain(self.scene)
        self.balls = [soya.Body(self.scene) for i in xrange(5)]
        self.balls_out = [soya.Body(self.scene) for i in xrange(4)]
        radius = 0.5
        for ball in self.balls:
            ball.mass = soya.SphericalMass(0.1, radius, u"total_mass")
            ball.y = 3
            ball.scale(radius, radius, radius)
            soya.GeomSphere(ball, radius)
        for ball in self.balls_out:
            ball.mass = soya.SphericalMass(0.1, radius, u"total_mass")
            ball.y = 3
            ball.scale(radius, radius, radius)
            soya.GeomSphere(ball, radius)
        self.main_loop = soya.MainLoop(self.scene)
        soya.init(quiet=True)

    def tearDown(self):
        """
        Delete created objects.
        """
        del self.scene, self.terrain

    def spend_time(self, time):
        for i in xrange(time):
            self.main_loop.begin_round()
            self.main_loop.advance_time(1)
            self.main_loop.end_round()

    def dispose_balls(self):
        "Dispose the balls upon the terrain, and balls_out aside it."
        x = -0.8*self.terrain.width/2
        z = -0.8*self.terrain.length/2
        for ball in self.balls:
            ball.x = self.terrain.x + x
            ball.z = self.terrain.z + z
            x += 0.2*self.terrain.width
            z += 0.2*self.terrain.length
        self.balls_out[0].x = self.terrain.x - self.terrain.width
        self.balls_out[0].z = self.terrain.z
        self.balls_out[1].x = self.terrain.x + self.terrain.width
        self.balls_out[1].z = self.terrain.z
        self.balls_out[2].x = self.terrain.x
        self.balls_out[2].z = self.terrain.z + self.terrain.length
        self.balls_out[3].x = self.terrain.x
        self.balls_out[3].z = self.terrain.z - self.terrain.length

    def test_ground(self):
        """
        Test the altitude of thrown balls upon and aside the terrain.
        """
        self.dispose_balls()

        self.spend_time(30)

        for ball in self.balls:
            self.assertTrue(ball.y > 0, "ball.y = %s" % ball.y)
        for ball in self.balls_out:
            self.assertTrue(ball.y < 0, "ball.y = %s" % ball.y)

    def test_walls(self):
        """
        Test wether the balls are blocked by the walls.
        """
        self.dispose_balls()

        for ball in self.balls:
            ball.y = self.terrain.wall_height
            ball.add_force(soya.Vector(ball,100,0,0))

        self.spend_time(20)

        for ball in self.balls:
            self.assertTrue(-self.terrain.width/2 <= self.terrain.x + ball.x <=
                    self.terrain.width/2 
                    and
                    -self.terrain.length/2 <= self.terrain.z + ball.z <=
                    self.terrain.length/2, "ball.x, ball.z = %s, %s" % (ball.x,
                    ball.z))

    def test_middle_wall(self):
        """
        Test wether the ball is blocked by the middle wall.
        """
        self.dispose_balls()

        ball = self.balls[0]
        ball.y = self.terrain.wall_height
        ball.x = self.terrain.width/2 - self.terrain.offside/2
        ball.z = -self.terrain.length/4
        ball.add_force(soya.Vector(ball,0,0,100))

        self.spend_time(20)

        self.assertTrue(ball.z < 0, u"ball.z = %s" % ball.z)

    def test_placement(self):
        """
        If the terrain is moved, its geometry is, too.

        A ball thrown upon it must hav a y coord positive, negative otherwise.
        """
        self.terrain.x = 25
        self.dispose_balls()

        self.spend_time(30)

        for ball in self.balls:
            self.assertTrue(ball.y > 0, u"ball.y = %s" % ball.y)
        for ball in self.balls_out:
            self.assertTrue(ball.y < 0, u"ball.y = %s" % ball.y)


class Terrain_parameters_TestCase(TestCase):
    """
    Tests on the change of dimensions of the terrains.
    """
    def setUp(self):
        """
        Initilize soya.
        """
        # Initilazing soya
        import soya
        global soya
        soya.path.append(os.path.join(os.path.dirname(sys.argv[0]), "../data"))
        self.scene = soya.World()  # Create the scene

    def tearDown(self):
        """
        Delete created objects.
        """
        del self.scene

    def test_constructor_dimension(self):
        """
        Test passing dimension parameters to the constructor.
        """
        width = 20
        length = 50
        wall_height = 5
        offside = 3
        border = 1
        middle_wall = 5
        
        terrain = models.Terrain(self.scene, autobuild=True, width=width,
                length=length, wall_height=wall_height, offside=offside,
                border=border, middle_wall=middle_wall)

        self.assertEqual(terrain.width, width)
        self.assertEqual(terrain.length, length)
        self.assertEqual(terrain.wall_height, wall_height)
        self.assertEqual(terrain.offside, offside)
        self.assertEqual(terrain.border, border)
        self.assertEqual(terrain.middle_wall, middle_wall)

    def test_build_dimension(self):
        """
        Test manual building of the terrain
        """
        width = 20
        length = 50
        wall_height = 5
        offside = 3
        border = 1
        middle_wall = 5
        
        terrain = models.Terrain(self.scene, autobuild=False)

        # Since autobuild=False, there must be no model.
        self.assertTrue(terrain.model is None)

        terrain.width = width
        terrain.length = length
        terrain.wall_height = wall_height
        terrain.offside = offside
        terrain.border = border
        terrain.middle_wall = middle_wall

        # Let's build the model
        terrain.build_model()

        self.assertTrue(terrain.model.nb_vertices > 0)

        total_box = terrain.get_box()
        total_width = abs(total_box[0].x - total_box[1].x)
        total_length = abs(total_box[0].z - total_box[1].z)
        total_height = abs(total_box[0].y - total_box[1].y)

        self.assertTrue(total_width >= width)
        self.assertTrue(total_length >= length)
        self.assertTrue(total_height >= wall_height)


class Ball_TestCase(TestCase):
    """
    Simply test the Ball object.
    """

    def setUp(self):
        """
        Create the scene.
        """
        import soya
        global soya
        self.scene = soya.World()

    def tearDown(self):
        """
        Delete the scene.
        """
        del self.scene

    def test_new_Ball(self):
        """
        Test the creation of a new Ball.
        """
        x = 3
        y = 5
        z = 17
        radius = 4
        material = soya.Material()
        material.diffuse = 1,0,0,1 # Red

        ball = models.Ball(self.scene, x, y, z, radius, material)

        # Get the wrapping box to test its size
        box = ball.get_box()
        width = abs(box[0].x - box[1].x)
        height = abs(box[0].y - box[1].y)
        length = abs(box[0].z - box[1].z)
        # Notvery strict tests on dimensions : since the ball is not a perfect
        # sphere, we must accept large difference between expected radius and
        # real one.
        self.assertTrue(radius <= width <= 2 * radius)
        self.assertEqual(2 * radius, height)
        self.assertTrue(radius <= length <= 2 * radius)
        
        self.assertEqual(ball.x, x)
        self.assertEqual(ball.y, y)
        self.assertEqual(ball.z, z)

        self.assertEqual(ball.material, material)


class Khepera_TestCase(TestCase):
    """
    Test the geometry of the Khepera object.
    """

    def setUp(self):
        """
        Create the scene.
        """
        import soya
        global soya
        self.scene = soya.World()

    def tearDown(self):
        """
        Delete the scene.
        """
        del self.scene

    def test_new_Khepera(self):
        """
        Test the creation of a new Khepera.
        """
        x = 3
        y = 5
        z = 17
        material = soya.Material()
        material.diffuse = 1,0,0,1 # Red

        khepera = models.Khepera(self.scene, True, x, y, z, material)

        # Get the wrapping box to test its size
        box = khepera.get_box()
        width = abs(box[0].x - box[1].x)
        heigth = abs(box[0].y - box[1].y)
        length = abs(box[0].z - box[1].z)
        self.assertAlmostEqual(max(2 * khepera.body_radius, 2 *
            khepera.wheel_offcenter + khepera.wheel_width), length, 3)
        self.assertAlmostEqual(khepera.body_height + khepera.wheel_radius -
                khepera.wheel_alt + khepera.led_radius, heigth, 3)
        self.assertAlmostEqual(2 * khepera.body_radius, width, 3)
        
        self.assertEqual(khepera.x, x)
        self.assertEqual(khepera.y, y)
        self.assertEqual(khepera.z, z)

        self.assertEqual(khepera.symbol_material, material)

    def test_no_order(self):
        """
        Test the no treatment of a new order without begin_roung
        """
        left_speed = 0.4
        right_speed = -0.6

        khepera = models.Khepera(self.scene)

        # Send the order
        khepera.new_order(left_speed,right_speed)
        (l_before, r_before) = khepera.get_speed()
        # Before 'begin_round' the speed doesn't change...
        self.assertEqual(l_before,0)
        self.assertEqual(r_before,0)
        
    def test_new_order(self):
        """
        Test the treatment of a new order
        """
        left_speed = 0.4
        right_speed = -0.6

        khepera = models.Khepera(self.scene)

        # Send the order
        khepera.new_order(left_speed,right_speed)
        
        # Treat the order
        khepera.read_order()
        (l, r) = khepera.get_speed()
        # ...after it does
        self.assertEqual(l,left_speed)
        self.assertEqual(r,right_speed)

    def test_bad_order(self):
        """
        Test the treatment of a new bad order
        """
        left_speed = "left"
        right_speed = "right"

        khepera = models.Khepera(self.scene)

        # Send the order
        khepera.new_order(left_speed,right_speed)
        try:
            khepera.read_order()
        except TypeError:
            self.assertTrue(True)
        else:
            self.assertFalse(True)
        # TODO:ça ne marche pas, problème d'appel de la fonction...
        #self.assertRaises(self, ValueError, khepera.read_order, khepera)



class TravelingCamera_TestCase(TestCase):
    """
    Test the trajectory of the traveling camera, specialy outside of the field.
    """
    def setUp(self):
        import soya
        global soya

    def test_Camera_moving_outside(self):
        """
        Test the trajectory of the camera, outside of the field, when it starts
        parallele of the field's border.
        """
        # dimensions of the field
        width = float(4)
        length = float(6)
        # parameters of the move
        self.delta = float(1)
        self.distance = float(1)
        # initial position of the camera
        x_init = width/2 + self.distance
        y_init = length/2
        z_init = 0

        camera = models.TravelingCamera(soya.World(), width, length, delta=self.delta)
        camera.set_xyz(x_init, z_init, y_init)
        self.camera = camera
        
        self._turn_around(length, width, x_init, y_init)


    def test_Camera_moving_inside(self):
        """
        Test the trajectory of the camera, inside the field, when it starts
        parallele of the field's border.
        """
        # dimensions of the field
        field_width  = float(4)
        field_length = float(6)
        # parameters of the move
        self.delta = float(1)
        # real distance between the camera and the virtual borders
        self.distance = float(0.01)
        #dimensions of the virtual field
        width  = (field_width/2  - 1 - self.distance)*2
        length = (field_length/2 - 1 - self.distance)*2
        # initial position of the camera
        x_init = width/2 + self.distance
        y_init = length/2
        z_init = 0

        camera = models.TravelingCamera(soya.World(), field_width, field_length, delta=self.delta)
        camera.set_xyz(x_init, z_init, y_init)
        self.camera = camera
        
        self._turn_around(length, width, x_init, y_init)


    def _turn_around(self, length, width, x_init, y_init):
        #TODO: after the good coordinates change, be careful of the direction
        #(translate right or left?)
        self._round_border(length, x_init, -length/2)
        self._round_corner(width/2, -(length/2+self.distance))

        self._round_border(width, -width/2, -(length/2+self.distance))
        self._round_corner(-(width/2+self.distance), -length/2)

        self._round_border(length, -(width/2+self.distance), length/2)
        self._round_corner(-width/2, length/2+self.distance)

        self._round_border(width, width/2, length/2+self.distance)
        self._round_corner(x_init, y_init)

    def _round_range(self, round_range, x_value, y_value):
        if len(round_range)==0:
            return
        for i in round_range:
            self.camera.translate_right()
            pos = self.camera.position()
            (x, y) = pos.x, pos.z
        #Almost because of some irregular approximation with floats in soya
        self.assertAlmostEqual(x, x_value, 4)
        self.assertAlmostEqual(y, y_value, 4)

    def _round_corner(self, x_value, y_value):
        # we keep the int part, otherwise the camera goes too far
        round_range = range(int(self.distance*2 / self.delta))
        self._round_range(round_range, x_value, y_value)

    def _round_border(self, side, x_value, y_value):
        # '+1' because at a corner, the distance is div by 2
        round_range = range(int(math.ceil(side / self.delta)) + 1)
        self._round_range(round_range, x_value, y_value)


class Cylinder_TestCase(TestCase):
    """
    Test the geometry of the cylinder.
    """
    def setUp(self):
        import soya
        global soya

    def test_new_Cylinder(self):
        """
        Test the creation of a new cylinder.
        """
        slices = 20
        radius = 12
        length = 52

        world = cylinder.Cylinder(None, length, radius, slices=slices)

        # Get the wrapping box to test its size
        box = world.get_box()
        real_width = abs(box[0].x - box[1].x)
        real_heigth = abs(box[0].y - box[1].y)
        real_length = abs(box[0].z - box[1].z)
        self.assertAlmostEqual(2 * radius, real_width)
        self.assertAlmostEqual(2 * radius, real_heigth)
        self.assertEqual(length, real_length)


class Screenshoter_TestCase(SoyaTestCase):
    """
    Test the Screenshoter class.
    """
    def test_screenshot(self):
        """
        Test wether screenshot() let the camera unchanged and change img.
        """
        screenshoter = models.Screenshoter(self.scene)

        img, _ = screenshoter.get_image(time_out=0)
        self.assertEqual(img, None)

        screenshoter.screenshot()
        self.assertEqual(soya.root_widget, self.camera)
        img, _ = screenshoter.get_image()
        self.failIfEqual(img, None)

        screenshoter.screenshot()
        img_after, _ = screenshoter.get_image()
        self.failIfEqual(img, img_after)

    def test_update_screenshot(self):
        """
        Test wether update_screenshot() change the img every period.
        """
        screenshoter = models.Screenshoter(self.scene)
        screenshoter.period = period = 5

        for i in xrange(period - 1):
            screenshoter.update_screenshot()
            self.assertEqual(screenshoter.img, None)
        screenshoter.update_screenshot()
        img, _ = screenshoter.get_image()
        self.failIfEqual(img, None)
        for i in xrange(period - 1):
            screenshoter.update_screenshot()
            img_after, _ = screenshoter.get_image()
            self.assertEqual(img, img_after)
            img = img_after
        screenshoter.update_screenshot()
        img_after, _ = screenshoter.get_image()
        self.failIfEqual(img, img_after)





if __name__ == '__main__':
    unittest_main()
