#!/usr/bin/env python
# -*- coding:utf-8 -*-

from unittest import TestCase, main as unittest_main
import sys
import os
from math import degrees, hypot
import time

from threading import Thread
import soya
import Pyro.core
import PIL.Image

from scor.sonere.simulator import DistantSimulator, debug
from scor.sonere.models import Khepera, Ball
from scor.sonere.interfaces import Coordinate, Control, KheperaSerial, Video


MOVE_MASK = 'D,%i,%i\n'

class Pyro_TestCase(TestCase):
    """
    Test the pyro server of the simulator.
    """

    testNumber=0
    sim = None
        
    # Parameter the move
    SPEED = 2
    TICK_SPEED = SPEED / Khepera.PULSE_SPEED_RATIO
    NB_ROUND = 10
    # TODO: calcule the min borne of distance depending of ticks stuff
    MIN_DISTANCE = NB_ROUND * SPEED * 0.005

    def setUp(self):
        Pyro_TestCase.testNumber += 1
        if Pyro_TestCase.sim==None:
            soya.init(quiet=True)
            Pyro_TestCase.sim = DistantSimulator()
            self.sim.start_pyro()

    def tearDown(self):
        if Pyro_TestCase.testNumber==3:
            self.sim.stop_pyro()


    def get_id(self):
        return self.sim.config.robots_ids()[0]

    def test_coordinates(self):
        """
        Test the results provided by Pyro Coordinate.
        """
        # Connect to pyro server
        host = self.sim.config.get("pyro_server","coordinate_host")
        port = self.sim.config.get("pyro_server","coordinate_port")
        uri = u"PYROLOC://%s:%s/coordinate/" % (host, port)
        # Get proxy for coordinate
        coord = Pyro.core.getProxyForURI(uri+u'south')
        coord2 = Pyro.core.getProxyForURI(uri+u'north')
        # Get real coordinate from robots
        robots =  self.sim.config.robots_ids()
        robots_coord = {}
        for id_robot in robots:
            # TODO: replace '-z' by 'y' when coordinate will be good
            robots_coord[id_robot] = self.sim.kheperas[id_robot].get_coord()
        # Update coordinate informations
        coord.update_info()
        coord2.update_info()
        # Get coordinate from pyro proxy
        infos = coord.get_info(time_out=2)
        for id_robot in robots:
            x = robots_coord[id_robot][0]
            y = robots_coord[id_robot][1]
            self.assertAlmostEqual(infos[u'khep_coord'][id_robot][0], x, 1)
            self.assertAlmostEqual(infos[u'khep_coord'][id_robot][1], y, 1)

    def test_control(self):
        """
        Test Pyro Control calls
        """
        # Connect to pyro server
        host = self.sim.config.get("pyro_server","control_host")
        port = self.sim.config.get("pyro_server","control_port")
        uri = u"PYROLOC://%s:%s/control" % (host, port)
        # Get proxy for control
        control = Pyro.core.getProxyForURI(uri)
        # Get actual robot coordinate
        robot = self.sim.kheperas[self.get_id()]
        x, y, _ = robot.get_coord()
        # Move the robot and stop the khepera, for the next tests
        control.move(self.get_id(), Pyro_TestCase.SPEED)
        debug(self.sim.scene, Pyro_TestCase.NB_ROUND)
        control.stop(self.get_id())
        debug(self.sim.scene, 1)
        # Get new robot coordinate, and reinit coordinates
        x_after, y_after, _ = robot.get_coord()
        self.sim.kheperas[self.get_id()].init_coordinates()
        # Calcul the distance of the move
        distance =  hypot(x_after-x, y_after-y)
        self.assertTrue(distance > Pyro_TestCase.MIN_DISTANCE)

    def test_serial(self):
        """
        Test Pyro Serial calls
        """
        # Connect to pyro server
        host = self.sim.config.get("pyro_server","serial_host")
        port = self.sim.config.get("pyro_server","serial_port")
        port_khep = self.sim.config.get_robot_infos(self.get_id())[u'port']
        uri = u"PYROLOC://%s:%s/serial/%s" % (host, port, port_khep)
        # Get proxy for control
        serial = Pyro.core.getProxyForURI(uri)
        # Get actual robot coordinate
        robot = self.sim.kheperas[self.get_id()]
        x, y, _ = robot.get_coord()
        # Move the robot and stop the khepera, for the next tests
        serial.write(MOVE_MASK % (Pyro_TestCase.TICK_SPEED, Pyro_TestCase.TICK_SPEED))
        debug(self.sim.scene, Pyro_TestCase.NB_ROUND)
        serial.write(MOVE_MASK % (0,0))
        debug(self.sim.scene, 1)
        # Get new robot coordinate, and reinit coordinates
        x_after, y_after, _ = robot.get_coord()
        self.sim.kheperas[self.get_id()].init_coordinates()
        # Calcul the distance of the move
        distance =  hypot(x_after-x, y_after-y)
        self.assertTrue(distance > Pyro_TestCase.MIN_DISTANCE)


if __name__ == '__main__':
    unittest_main()
