# -*- coding:utf-8 -*-
"""Module providing useful class for the wheel of the robot."""

from soya import HingeJoint, AngularMotor, Point

class MotoredAxleJoint(object):
    """Functional class to handle a motored joint.
    
    Include a hinge joint and a angular motor, which work together to give a
    practical model of a wheel."""
    def __init__(self, vehicule, wheel, axis):
        """
        Initialize the hinge joint and the angular motor.
        """
        self._vehicule = vehicule
        self._wheel = wheel
        self._axis = axis
        self._reversed = False
        self._speed = 0

        self._hinge = HingeJoint(self.vehicule, self.wheel)
        self._hinge.anchor = Point(self.wheel)
        self._hinge.axis  = self.axis

        self._motor = AngularMotor(self.vehicule, self.wheel)
        self._motor.nb_axes = 1
        self._motor.setAxis(0, 2, axis)

        self._hinge.cfm = 10000
        self._hinge.bounce = 0
        self._hinge.fudge_factor = 1000
        self._motor.fmax =  1

    @property
    def vehicule(self):
        """
        Vehicule of the Joint.
        """
        return self._vehicule

    @property
    def wheel(self):
        """
        Wheel of the Joint.
        """
        return self._wheel

    @property
    def axis(self):
        """
        Axis of the Joint.
        """
        return self._axis

    def _set_angular_speed(self, speed):
        """
        Set the angular speed by given speed parameter.
        """
        #self._motor.lo_stop = -self._motor.hi_stop # XXX
        self._speed = speed
        if speed < 0:
            if not self._reversed:
                self._motor.setAxis(0, 2, -self.axis)
                self._reversed = True
            speed = abs(speed)
        elif self._reversed:
            self._motor.setAxis(0, 2, self.axis)
            self._reversed = False
        self._motor.velocity = speed

    def _get_angular_speed(self):
        """
        Return the speed.
        """
        return self._speed

    angular_speed = property(fget=_get_angular_speed, fset=_set_angular_speed,
                             doc="the requested angular speed of the motor")

    def _set_strength(self, strenght):
        """
        Change the strength of the motor.
        """
        self._motor.fmax = strenght

    def _get_strength(self):
        """
        Return the strength of the motor
        """
        return self._motor.fmax

    strength = property(fget=_get_strength, fset=_set_strength,
                        doc="the strength used to achieved this velocity")


